#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test suite for the StationXML reader and writer.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2013
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import fnmatch
import io
import re
import warnings

from lxml import etree

import obspy
import obspy.io.stationxml.core
from obspy import UTCDateTime
from obspy.core.util import AttribDict, CatchAndAssertWarnings
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning
from obspy.core.inventory import (
    Inventory, Network, ResponseStage, Response, Channel, Station)
from obspy.core.inventory.util import DataAvailability
from obspy.core.util.base import NamedTemporaryFile
from obspy.io.stationxml.core import _read_stationxml
import pytest


class TestStationXML():
    """
    """
    @classmethod
    def setup_class(cls):
        cls.maxDiff = 10000

    def _assert_station_xml_equality(self, xml_file_buffer,
                                     expected_xml_file_buffer):
        """
        Helper function comparing two BytesIO buffers contain Station XML
        files.
        """
        new_lines = [_i.decode('utf-8').strip().replace("'", '"')
                     for _i in xml_file_buffer.read().splitlines()]
        org_lines = [_i.decode('utf-8').strip().replace("'", '"')
                     for _i in expected_xml_file_buffer.read().splitlines()]

        # Remove the module lines from the original file.
        org_lines = [_i.strip() for _i in org_lines
                     if not _i.strip().startswith("<Module")]

        for new_line, org_line in zip(new_lines, org_lines):
            regex = "<(.*?) (.*?)/?>"

            def callback(pattern):
                part2 = " ".join(sorted(pattern.group(2).split(" ")))
                return "<%s %s>" % (pattern.group(1), part2)

            # resort attributes alphabetically
            org_line = re.sub(regex, callback, org_line, count=1)
            new_line = re.sub(regex, callback, new_line, count=1)
            assert org_line == new_line

        # Assert the line length at the end to find trailing non-equal lines.
        # If it is done before the line comparison it is oftentimes not very
        # helpful as you do not know which line is missing.
        assert len(new_lines) == len(org_lines)

    def test_is_stationxml(self, testdata, datapath):
        """
        Tests the _is_stationxml() function.
        """
        # Check positives.
        stationxmls = [testdata["minimal_station.xml"]]
        for stat in stationxmls:
            assert obspy.io.stationxml.core._is_stationxml(stat)

        # Check some negatives.
        not_stationxmls = [
            "Variations-FDSNSXML-SEED.txt",
            "fdsn-station+availability-1.0.xsd", "fdsn-station-1.0.xsd"]
        not_stationxmls = [datapath.parent.parent / 'data' / name
                           for name in not_stationxmls]
        for stat in not_stationxmls:
            assert not obspy.io.stationxml.core._is_stationxml(stat)

    def test_different_write_levels(self, testdata):
        """
        Tests different levels of writing
        """
        filename = testdata["stationxml_BK.CMB.__.LKS.xml"]
        inv = obspy.read_inventory(filename)

        # Write to network level
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", level="network")
        file_buffer.seek(0, 0)

        network_inv = obspy.read_inventory(file_buffer)

        assert len(network_inv.networks) == len(inv.networks)

        for net in network_inv.networks:
            assert len(net.stations) == 0

        # Write to station level
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", level="station")
        file_buffer.seek(0, 0)

        station_inv = obspy.read_inventory(file_buffer)

        for net in station_inv.networks:
            assert len(net.stations) == len(inv[0].stations)
            for sta in net.stations:
                assert len(sta.channels) == 0

        # Write to channel level
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", level="channel")
        file_buffer.seek(0, 0)

        channel_inv = obspy.read_inventory(file_buffer)

        for net in channel_inv.networks:
            assert len(net.stations) == len(inv[0].stations)
            for sta in net.stations:
                assert len(sta.channels) == len(inv[0][0].channels)
                for cha in sta.channels:
                    assert cha.response is None

    def test_read_and_write_minimal_file(self, testdata):
        """
        Test that writing the most basic StationXML document possible works.
        """
        filename = testdata["minimal_station.xml"]
        inv = obspy.read_inventory(filename)

        # Assert the few values that are set directly.
        assert inv.source == "OBS"
        assert inv.created == obspy.UTCDateTime(2013, 1, 1)
        assert len(inv.networks) == 1
        assert inv.networks[0].code == "PY"

        # Write it again. Also validate it to get more confidence. Suppress the
        # writing of the ObsPy related tags to ease testing.
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", validate=True,
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(file_buffer,
                                          expected_xml_file_buffer)

    def test_subsecond_read_and_write_minimal_file(self, testdata):
        """
        Test reading and writing of sub-second time in datetime field,
        using creation time

        """
        filename = testdata["minimal_station_with_microseconds.xml"]
        inv = obspy.read_inventory(filename)

        # Write it again. Also validate it to get more confidence. Suppress the
        # writing of the ObsPy related tags to ease testing.
        file_buffer = io.BytesIO()

        inv.write(file_buffer, format="StationXML", validate=True,
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(file_buffer,
                                          expected_xml_file_buffer)

    def test_read_and_write_full_file(self, testdata):
        """
        Test that reading and writing of a full StationXML document with all
        possible tags works.
        """
        filename = testdata["full_random_stationxml.xml"]
        inv = obspy.read_inventory(filename)

        # Write it again. Also validate it to get more confidence. Suppress the
        # writing of the ObsPy related tags to ease testing.
        file_buffer = io.BytesIO()

        inv.write(file_buffer, format="StationXML", validate=True,
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(file_buffer,
                                          expected_xml_file_buffer)

        # test some new fields added in StationXML 1.1 specifically
        net = inv[0]
        sta = net[0]
        cha = sta[0]
        assert len(net.identifiers) == 2
        assert net.identifiers[0] == "abc:def"
        assert net.identifiers[1] == "uvw:xyz"
        assert len(net.operators) == 2
        assert net.operators[0].agency == 'ABC0.oszQNsC4l66ieQFM'
        # accessing "agencies" this should raise a DeprecationWarning but its
        # tested in another test case already
        with CatchAndAssertWarnings():
            assert net.operators[0].agencies == \
                             ['ABC0.oszQNsC4l66ieQFM']
        assert net.operators[0].contacts[0].names[0] == 'A'
        # check WaterLevel tag
        assert sta.water_level == 250.4
        assert sta.water_level.lower_uncertainty == 2.3
        assert sta.water_level.upper_uncertainty == 4.2
        assert sta.water_level.unit == 'METERS'
        assert cha.water_level == 631.2
        assert cha.water_level.lower_uncertainty == 5.3
        assert cha.water_level.upper_uncertainty == 3.2
        assert cha.water_level.unit == 'METERS'
        # multiple equipments allowed now on channel, deprecation for old
        # single equipment attribute
        assert len(cha.equipments) == 2
        assert cha.equipments[1].type == "some type"
        msg = (r"Attribute 'equipment' \(holding a single Equipment\) is "
               r"deprecated in favor of 'equipments' which now holds a list "
               r"of Equipment objects \(following changes in StationXML 1.1\) "
               r"and might be removed in the future. Returning the first "
               r"entry found in 'equipments'.")
        with CatchAndAssertWarnings(
                clear=['obspy.core.inventory.channel'],
                expected=[(ObsPyDeprecationWarning, msg)]):
            assert cha.equipment.type == cha.equipments[0].type
        # check new measurementMethod attributes
        assert sta.latitude.measurement_method == "GPS"
        assert sta.longitude.measurement_method == "GPS"
        assert sta.elevation.measurement_method == "digital elevation model"
        assert cha.azimuth.measurement_method == "fibre optic gyro compass"
        # check data availability tags
        assert net.data_availability.start == UTCDateTime(2011, 2, 3, 4, 5, 6)
        assert net.data_availability.end == UTCDateTime(2011, 3, 4, 5, 6, 7)
        assert len(net.data_availability.spans) == 2
        span1 = net.data_availability.spans[0]
        span2 = net.data_availability.spans[1]
        assert span1.start == UTCDateTime(2012, 2, 3, 4, 5, 6)
        assert span1.end == UTCDateTime(2012, 3, 4, 5, 6, 7)
        assert span1.number_of_segments == 5
        assert span1.maximum_time_tear == 7.8
        assert span2.start == UTCDateTime(2013, 2, 3, 4, 5, 6)
        assert span2.end == UTCDateTime(2013, 3, 4, 5, 6, 7)
        assert span2.number_of_segments == 8
        assert span2.maximum_time_tear == 2.4
        # test sourceID
        assert net.source_id == "http://www.example.com"
        assert sta.source_id == "http://www.example2.com"
        assert cha.source_id == "http://www.example3.com"
        # Comment topic
        assert net.comments[0].subject == "my topic"
        # Comment id optional now
        assert net.comments[0].id is None
        # storage_format was deprecated since it was removed in StationXML 1.1
        msg = (r"Attribute 'storage_format' was removed in accordance with "
               r"StationXML 1\.1, ignoring\.")
        with CatchAndAssertWarnings(
                clear=['obspy.core.inventory.channel'],
                expected=[(ObsPyDeprecationWarning, msg)]):
            cha.storage_format = "something"
        msg = (r"Attribute 'storage_format' was removed in accordance with "
               r"StationXML 1\.1, returning None\.")
        with CatchAndAssertWarnings(
                clear=['obspy.core.inventory.channel'],
                expected=[(ObsPyDeprecationWarning, msg)]):
            assert cha.storage_format is None
        # check new number attributes on Numerator/Denominator
        resp = cha.response
        stage = resp.response_stages[1]
        assert stage.numerator[0].number == 1
        assert stage.numerator[1].number == 2
        assert stage.denominator[0].number == 3
        assert stage.denominator[1].number == 4

    def test_writing_module_tags(self):
        """
        Tests the writing of ObsPy related tags.
        """
        net = Network(code="UL")
        inv = Inventory(networks=[net], source="BLU")

        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", validate=True)
        file_buffer.seek(0, 0)
        lines = file_buffer.read().decode().splitlines()
        module_line = [_i.strip() for _i in lines if _i.strip().startswith(
            "<Module>")][0]
        assert fnmatch.fnmatch(module_line, "<Module>ObsPy *</Module>")
        module_uri_line = [_i.strip() for _i in lines if
                           _i.strip().startswith("<ModuleURI>")][0]
        assert module_uri_line == \
            "<ModuleURI>https://www.obspy.org</ModuleURI>"

    def test_reading_other_module_tags(self, testdata):
        """
        Even though the ObsPy Tags are always written, other tags should be
        able to be read.
        """
        filename = testdata[
            "minimal_with_non_obspy_module_and_sender_tags_station.xml"]
        inv = obspy.read_inventory(filename)
        assert inv.module == "Some Random Module"
        assert inv.module_uri == "http://www.some-random.site"

    def test_reading_and_writing_full_root_tag(self, testdata):
        """
        Tests reading and writing a full StationXML root tag.
        """
        filename = testdata[
            "minimal_with_non_obspy_module_and_sender_tags_station.xml"]
        inv = obspy.read_inventory(filename)
        assert inv.source == "OBS"
        assert inv.created == obspy.UTCDateTime(2013, 1, 1)
        assert len(inv.networks) == 1
        assert inv.networks[0].code == "PY"
        assert inv.module == "Some Random Module"
        assert inv.module_uri == "http://www.some-random.site"
        assert inv.sender == "The ObsPy Team"

        # Write it again. Do not write the module tags.
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", validate=True,
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(
            file_buffer, expected_xml_file_buffer)

    def test_reading_and_writing_full_network_tag(self, testdata):
        """
        Tests the reading and writing of a file with a more or less full
        network tag.
        """
        filename = testdata["full_network_field_station.xml"]
        inv = obspy.read_inventory(filename)

        # Assert all the values...
        assert len(inv.networks) == 1
        net = inv.networks[0]
        assert net.code == "PY"
        assert net.start_date == obspy.UTCDateTime(2011, 1, 1)
        assert net.end_date == obspy.UTCDateTime(2012, 1, 1)
        assert net.restricted_status == "open"
        assert net.alternate_code == "PYY"
        assert net.historical_code == "YYP"
        assert net.description == "Some Description..."
        assert len(net.comments) == 2

        comment_1 = net.comments[0]
        assert comment_1.value == "Comment number 1"
        assert comment_1.begin_effective_time == obspy.UTCDateTime(1990, 5, 5)
        assert comment_1.end_effective_time == obspy.UTCDateTime(2008, 2, 3)
        assert len(comment_1.authors) == 1
        authors = comment_1.authors[0]
        assert len(authors.names) == 2
        assert authors.names[0] == "This person"
        assert authors.names[1] == "has multiple names!"
        assert len(authors.agencies) == 3
        assert authors.agencies[0] == "And also"
        assert authors.agencies[1] == "many"
        assert authors.agencies[2] == "many Agencies"
        assert len(authors.emails) == 4
        assert authors.emails[0] == "email1@mail.com"
        assert authors.emails[1] == "email2@mail.com"
        assert authors.emails[2] == "email3@mail.com"
        assert authors.emails[3] == "email4@mail.com"
        assert len(authors.phones) == 2
        assert authors.phones[0].description == "phone number 1"
        assert authors.phones[0].country_code == 49
        assert authors.phones[0].area_code == 123
        assert authors.phones[0].phone_number == "456-7890"
        assert authors.phones[1].description == "phone number 2"
        assert authors.phones[1].country_code == 34
        assert authors.phones[1].area_code == 321
        assert authors.phones[1].phone_number == "129-7890"

        comment_2 = net.comments[1]
        assert comment_2.value == "Comment number 2"
        assert comment_2.begin_effective_time == obspy.UTCDateTime(1990, 5, 5)
        assert comment_1.end_effective_time == obspy.UTCDateTime(2008, 2, 3)
        assert len(comment_2.authors) == 3
        for _i, author in enumerate(comment_2.authors):
            assert len(author.names) == 1
            assert author.names[0] == "Person %i" % (_i + 1)
            assert len(author.agencies) == 1
            assert author.agencies[0] == "Some agency"
            assert len(author.emails) == 1
            assert author.emails[0] == "email@mail.com"
            assert len(author.phones) == 1
            assert author.phones[0].description is None
            assert author.phones[0].country_code == 49
            assert author.phones[0].area_code == 123
            assert author.phones[0].phone_number == "456-7890"

        # Now write it again and compare to the original file.
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", validate=True,
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(
            file_buffer,
            expected_xml_file_buffer)

    def test_reading_and_writing_full_station_tag(self, testdata):
        """
        Tests the reading and writing of a file with a more or less full
        station tag.
        """
        filename = testdata["full_station_field_station.xml"]
        inv = obspy.read_inventory(filename)

        # Assert all the values...
        assert len(inv.networks) == 1
        assert inv.source == "OBS"
        assert inv.module == "Some Random Module"
        assert inv.module_uri == "http://www.some-random.site"
        assert inv.sender == "The ObsPy Team"
        assert inv.created == obspy.UTCDateTime(2013, 1, 1)
        assert len(inv.networks) == 1
        network = inv.networks[0]
        assert network.code == "PY"

        # Now assert the station specific values.
        assert len(network.stations) == 1
        station = network.stations[0]
        assert station.code == "PY"
        assert station.start_date == obspy.UTCDateTime(2011, 1, 1)
        assert station.end_date == obspy.UTCDateTime(2012, 1, 1)
        assert station.restricted_status == "open"
        assert station.alternate_code == "PYY"
        assert station.historical_code == "YYP"
        assert station.description == "Some Description..."
        assert len(station.comments) == 2
        comment_1 = station.comments[0]
        assert comment_1.value == "Comment number 1"
        assert comment_1.begin_effective_time == obspy.UTCDateTime(1990, 5, 5)
        assert comment_1.end_effective_time == obspy.UTCDateTime(2008, 2, 3)
        assert len(comment_1.authors) == 1
        authors = comment_1.authors[0]
        assert len(authors.names) == 2
        assert authors.names[0] == "This person"
        assert authors.names[1] == "has multiple names!"
        assert len(authors.agencies) == 3
        assert authors.agencies[0] == "And also"
        assert authors.agencies[1] == "many"
        assert authors.agencies[2] == "many Agencies"
        assert len(authors.emails) == 4
        assert authors.emails[0] == "email1@mail.com"
        assert authors.emails[1] == "email2@mail.com"
        assert authors.emails[2] == "email3@mail.com"
        assert authors.emails[3] == "email4@mail.com"
        assert len(authors.phones) == 2
        assert authors.phones[0].description == "phone number 1"
        assert authors.phones[0].country_code == 49
        assert authors.phones[0].area_code == 123
        assert authors.phones[0].phone_number == "456-7890"
        assert authors.phones[1].description == "phone number 2"
        assert authors.phones[1].country_code == 34
        assert authors.phones[1].area_code == 321
        assert authors.phones[1].phone_number == "129-7890"
        comment_2 = station.comments[1]
        assert comment_2.value == "Comment number 2"
        assert comment_2.begin_effective_time == obspy.UTCDateTime(1990, 5, 5)
        assert comment_1.end_effective_time == obspy.UTCDateTime(2008, 2, 3)
        assert len(comment_2.authors) == 3
        for _i, author in enumerate(comment_2.authors):
            assert len(author.names) == 1
            assert author.names[0] == "Person %i" % (_i + 1)
            assert len(author.agencies) == 1
            assert author.agencies[0] == "Some agency"
            assert len(author.emails) == 1
            assert author.emails[0] == "email@mail.com"
            assert len(author.phones) == 1
            assert author.phones[0].description is None
            assert author.phones[0].country_code == 49
            assert author.phones[0].area_code == 123
            assert author.phones[0].phone_number == "456-7890"

        assert station.latitude == 10.0
        assert station.longitude == 20.0
        assert station.elevation == 100.0

        assert station.site.name == "Some site"
        assert station.site.description == "Some description"
        assert station.site.town == "Some town"
        assert station.site.county == "Some county"
        assert station.site.region == "Some region"
        assert station.site.country == "Some country"

        assert station.vault == "Some vault"
        assert station.geology == "Some geology"

        assert len(station.equipments) == 2
        assert station.equipments[0].resource_id == "some_id"
        assert station.equipments[0].type == "Some type"
        assert station.equipments[0].description == "Some description"
        assert station.equipments[0].manufacturer == "Some manufacturer"
        assert station.equipments[0].vendor == "Some vendor"
        assert station.equipments[0].model == "Some model"
        assert station.equipments[0].serial_number == "12345-ABC"
        assert station.equipments[0].installation_date == \
            obspy.UTCDateTime(1990, 5, 5)
        assert station.equipments[0].removal_date == \
            obspy.UTCDateTime(1999, 5, 5)
        assert station.equipments[0].calibration_dates[0] == \
            obspy.UTCDateTime(1990, 5, 5)
        assert station.equipments[0].calibration_dates[1] == \
            obspy.UTCDateTime(1992, 5, 5)
        assert station.equipments[1].resource_id == "something_new"
        assert station.equipments[1].type == "Some type"
        assert station.equipments[1].description == "Some description"
        assert station.equipments[1].manufacturer == "Some manufacturer"
        assert station.equipments[1].vendor == "Some vendor"
        assert station.equipments[1].model == "Some model"
        assert station.equipments[1].serial_number == "12345-ABC"
        assert station.equipments[1].installation_date == \
            obspy.UTCDateTime(1990, 5, 5)
        assert station.equipments[1].removal_date == \
            obspy.UTCDateTime(1999, 5, 5)
        assert station.equipments[1].calibration_dates[0] == \
            obspy.UTCDateTime(1990, 5, 5)
        assert station.equipments[1].calibration_dates[1] == \
            obspy.UTCDateTime(1992, 5, 5)

        assert len(station.operators) == 2
        assert station.operators[0].agency == "Agency 1"
        # legacy, "agencies" was a thing for StationXML 1.0
        regex = (
            r"Attribute 'agencies' \(holding a list of strings as Agencies\) "
            r"is deprecated in favor of 'agency' which now holds a single "
            r"string \(following changes in StationXML 1\.1\) and might be "
            r"removed in the future. Returning a list built up of the single "
            r"agency or an empty list if agency is None.")
        with CatchAndAssertWarnings(
                clear=['obspy.io.stationxml.core'],
                expected=[(ObsPyDeprecationWarning, regex)]):
            assert station.operators[0].agencies[0] == "Agency 1"
        assert station.operators[0].contacts[0].names[0] == "This person"
        assert station.operators[0].contacts[0].names[1] == \
            "has multiple names!"
        assert len(station.operators[0].contacts[0].agencies) == 3
        assert station.operators[0].contacts[0].agencies[0] == "And also"
        assert station.operators[0].contacts[0].agencies[1] == "many"
        assert station.operators[0].contacts[0].agencies[2] == "many Agencies"
        assert len(station.operators[0].contacts[0].emails) == 4
        assert station.operators[0].contacts[0].emails[0] == "email1@mail.com"
        assert station.operators[0].contacts[0].emails[1] == "email2@mail.com"
        assert station.operators[0].contacts[0].emails[2] == "email3@mail.com"
        assert station.operators[0].contacts[0].emails[3] == "email4@mail.com"
        assert len(station.operators[0].contacts[0].phones) == 2
        assert station.operators[0].contacts[0].phones[0].description == \
            "phone number 1"
        assert station.operators[0].contacts[0].phones[0].country_code == 49
        assert station.operators[0].contacts[0].phones[0].area_code == 123
        assert station.operators[0].contacts[0].phones[0].phone_number == \
            "456-7890"
        assert station.operators[0].contacts[0].phones[1].description == \
            "phone number 2"
        assert station.operators[0].contacts[0].phones[1].country_code == 34
        assert station.operators[0].contacts[0].phones[1].area_code == 321
        assert station.operators[0].contacts[0].phones[1].phone_number == \
            "129-7890"
        assert station.operators[0].contacts[1].names[0] == "Name"
        assert station.operators[0].contacts[1].agencies[0] == "Agency"
        assert station.operators[0].contacts[1].emails[0] == "email@mail.com"
        assert station.operators[0].contacts[1].phones[0].description == \
            "phone number 1"
        assert station.operators[0].contacts[1].phones[0].country_code == 49
        assert station.operators[0].contacts[1].phones[0].area_code == 123
        assert station.operators[0].contacts[1].phones[0].phone_number == \
            "456-7890"
        assert station.operators[0].website == "http://www.web.site"

        assert station.operators[1].agency == "Agency"
        assert station.operators[1].contacts[0].names[0] == "New Name"
        assert station.operators[1].contacts[0].agencies[0] == "Agency"
        assert station.operators[1].contacts[0].emails[0] == "email@mail.com"
        assert station.operators[1].contacts[0].phones[0].description == \
            "phone number 1"
        assert station.operators[1].contacts[0].phones[0].country_code == 49
        assert station.operators[1].contacts[0].phones[0].area_code == 123
        assert station.operators[1].contacts[0].phones[0].phone_number == \
            "456-7890"
        assert station.operators[1].website == "http://www.web.site"

        assert station.creation_date == obspy.UTCDateTime(1990, 5, 5)
        assert station.termination_date == obspy.UTCDateTime(2009, 5, 5)
        assert station.total_number_of_channels == 100
        assert station.selected_number_of_channels == 1

        assert len(station.external_references) == 2
        assert station.external_references[0].uri == \
            "http://path.to/something"
        assert station.external_references[0].description == \
            "Some description"
        assert station.external_references[1].uri == \
            "http://path.to/something/else"
        assert station.external_references[1].description == \
            "Some other description"

        # Now write it again and compare to the original file.
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML", validate=True,
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(file_buffer,
                                          expected_xml_file_buffer)

    def test_reading_and_writing_channel_with_response(self, testdata):
        """
        Test the reading and writing of a single channel including a
        multi-stage response object.
        """
        filename = testdata["IRIS_single_channel_with_response.xml"]
        inv = obspy.read_inventory(filename)
        assert inv.source == "IRIS-DMC"
        assert inv.sender == "IRIS-DMC"
        assert inv.created == obspy.UTCDateTime("2013-04-16T06:15:28")
        # Assert that precisely one channel object has been created.
        assert len(inv.networks) == 1
        assert len(inv.networks[0].stations) == 1
        assert len(inv.networks[0].stations[0].channels) == 1
        network = inv.networks[0]
        station = network.stations[0]
        channel = station.channels[0]
        # Assert some fields of the network. This is extensively tested
        # elsewhere.
        assert network.code == "IU"
        assert network.start_date == obspy.UTCDateTime("1988-01-01T00:00:00")
        assert network.end_date == obspy.UTCDateTime("2500-12-12T23:59:59")
        assert network.description == \
            "Global Seismograph Network (GSN - IRIS/USGS)"
        # Assert a few fields of the station. This is extensively tested
        # elsewhere.
        assert station.code == "ANMO"
        assert station.latitude == 34.94591
        assert station.longitude == -106.4572
        assert station.elevation == 1820.0
        assert station.site.name == "Albuquerque, New Mexico, USA"
        # Start to assert the channel reading.
        assert channel.code == "BHZ"
        assert channel.location_code == "10"
        assert channel.start_date == obspy.UTCDateTime("2012-03-13T08:10:00")
        assert channel.end_date == obspy.UTCDateTime("2599-12-31T23:59:59")
        assert channel.restricted_status == "open"
        assert channel.latitude == 34.945913
        assert channel.longitude == -106.457122
        assert channel.elevation == 1759.0
        assert channel.depth == 57.0
        assert channel.azimuth == 0.0
        assert channel.dip == -90.0
        assert channel.types == ["CONTINUOUS", "GEOPHYSICAL"]
        assert channel.sample_rate == 40.0
        assert channel.clock_drift_in_seconds_per_sample == 0.0
        assert channel.sensor.type == \
            "Guralp CMG3-T Seismometer (borehole)"
        # Check the response.
        response = channel.response
        sensitivity = response.instrument_sensitivity
        assert sensitivity.value == 3.31283E10
        assert sensitivity.frequency == 0.02
        assert sensitivity.input_units == "M/S"
        assert sensitivity.input_units_description == \
            "Velocity in Meters Per Second"
        assert sensitivity.output_units == "COUNTS"
        assert sensitivity.output_units_description == "Digital Counts"
        # Assert that there are three stages.
        assert len(response.response_stages) == 3

    def test_stationxml_with_availability(self, testdata):
        """
        A variant of StationXML has support for availability information.
        Make sure this works.
        """
        filename = testdata["stationxml_with_availability.xml"]
        inv = obspy.read_inventory(filename, format="stationxml")
        channel = inv[0][0][0]
        assert channel.data_availability.start == \
            obspy.UTCDateTime("1998-10-26T20:35:58")
        assert channel.data_availability.end == \
            obspy.UTCDateTime("2014-07-21T12:00:00")

        # Now write it again and compare to the original file.
        file_buffer = io.BytesIO()
        inv.write(file_buffer, format="StationXML",
                  _suppress_module_tags=True)
        file_buffer.seek(0, 0)

        with open(filename, "rb") as open_file:
            expected_xml_file_buffer = io.BytesIO(open_file.read())
        expected_xml_file_buffer.seek(0, 0)

        self._assert_station_xml_equality(file_buffer,
                                          expected_xml_file_buffer)

    def test_parse_file_with_no_default_namespace(self, testdata):
        """
        Tests that reading a file with no default namespace works fine.

        See #1060.
        """
        filename = testdata["no_default_namespace.xml"]
        inv = obspy.read_inventory(filename)
        # Very small file with almost no content.
        assert len(inv.networks) == 1
        assert inv[0].code == "XX"

    def test_parse_file_with_schema_2(self, testdata):
        """
        Reading a StationXML file version 2.0
        """
        filename = testdata["version20.xml"]

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            inv = obspy.read_inventory(filename)
        assert len(w) == 1
        assert 'StationXML file has version 2.0' in str(w[0].message)

        # Very small file with almost no content.
        assert len(inv.networks) == 1
        assert inv[0].code == "XX"

    def test_numbers_are_written_to_poles_and_zeros(self):
        """
        Poles and zeros have a number attribute. Make sure this is written,
        even if set with a custom complex list.
        """
        # Read default inventory and cut down to a single channel.
        inv = obspy.read_inventory()
        inv.networks = inv[:1]
        inv[0].stations = inv[0][:1]
        inv[0][0].channels = inv[0][0][:1]

        # Manually set the poles and zeros - thus these are cast to our
        # custom classes but number are not yet set.
        inv[0][0][0].response.response_stages[0].poles = [0 + 1j, 2 + 3j]
        inv[0][0][0].response.response_stages[0].zeros = [0 + 1j, 2 + 3j]

        with io.BytesIO() as buf:
            inv.write(buf, format="stationxml", validate=True)
            buf.seek(0, 0)
            data = buf.read().decode()

        # Ugly test - remove all whitespace and make sure the four following
        # lines are part of the written output.
        data = re.sub(r'\s+', ' ', data)

        assert '<Zero number="0"> <Real>0.0</Real> ' \
            '<Imaginary>1.0</Imaginary> </Zero>' in data
        assert '<Zero number="1"> <Real>2.0</Real> ' \
            '<Imaginary>3.0</Imaginary> </Zero>' in data
        assert '<Pole number="0"> <Real>0.0</Real> ' \
            '<Imaginary>1.0</Imaginary> </Pole>' in data
        assert '<Pole number="1"> <Real>2.0</Real> ' \
            '<Imaginary>3.0</Imaginary> </Pole>' in data

    def test_write_with_extra_tags_namespace_redef(self, testdata):
        """
        Tests the exceptions are raised when namespaces
        are redefined.
        """
        filename = testdata["stationxml_with_availability.xml"]
        # read the StationXML with availability
        inv = obspy.read_inventory(filename)
        with NamedTemporaryFile() as tf:
            # manually add custom namespace definition
            tmpfile = tf.name
            # assert that namespace prefix of None raises ValueError
            mynsmap = {None: 'http://bad.custom.ns/'}
            with pytest.raises(ValueError):
                inv.write(path_or_file_object=tmpfile,
                          format="STATIONXML", nsmap=mynsmap)

    def test_write_with_extra_tags_without_read_extra(self):
        """
        Tests that a Inventory object that was instantiated with
        custom namespace tags and attributes is written correctly.
        """
        # read the default inventory
        inv = obspy.read_inventory()
        # manually add extra to the dictionary
        network = inv[0]
        network.extra = {}
        ns = 'http://test.myns.ns/'
        # manually add a new custom namespace tag and attribute to the
        # inventory
        network.extra['mynsNetworkTag'] = AttribDict({
            'value': 'mynsNetworkTagValue',
            'namespace': ns})
        network.extra['mynsNetworkAttrib'] = AttribDict({
            'value': 'mynsNetworkAttribValue',
            'namespace': ns,
            'type': 'attribute'})
        station = inv[0][0]
        station.extra = {}
        station.extra['mynsStationTag'] = AttribDict({
            'value': 'mynsStationTagValue',
            'namespace': ns})
        station.extra['mynsStationAttrib'] = AttribDict({
            'value': 'mynsStationAttribValue',
            'namespace': ns,
            'type': 'attribute'})
        channel = inv[0][0][0]
        # add data availability to inventory
        channel.data_availability = DataAvailability(
            start=UTCDateTime('1998-10-26T20:35:58+00:00'),
            end=UTCDateTime('2014-07-21T12:00:00+00:00'))
        channel.extra = {}
        channel.extra['mynsChannelTag'] = AttribDict({
            'value': 'mynsChannelTagValue', 'namespace': ns})
        channel.extra['mynsChannelAttrib'] = AttribDict({
            'value': 'mynsChannelAttribValue',
            'namespace': ns,
            'type': 'attribute'})
        # add nested tags
        nested_tag = AttribDict()
        nested_tag.namespace = ns
        nested_tag.value = AttribDict()
        # add two nested tags
        nested_tag.value.my_nested_tag1 = AttribDict()
        nested_tag.value.my_nested_tag1.namespace = ns
        nested_tag.value.my_nested_tag1.value = 1.23E+10
        nested_tag.value.my_nested_tag2 = AttribDict()
        nested_tag.value.my_nested_tag2.namespace = ns
        nested_tag.value.my_nested_tag2.value = True
        nested_tag.value.my_nested_tag2.attrib = {'{%s}%s' % (
            ns, 'nestedAttribute1'): 'nestedAttributeValue1'}
        channel.extra['nested'] = nested_tag
        with NamedTemporaryFile() as tf:
            # manually add custom namespace definition
            tmpfile = tf.name
            # set namespace map to include only valid custom namespaces
            mynsmap = {'myns': ns}
            # write file with manually defined namespace map
            inv.write(tmpfile, format="STATIONXML", nsmap=mynsmap)
            # check contents
            with open(tmpfile, "rb") as fh:
                # enforce reproducible attribute orders through write_c14n
                obj = etree.fromstring(fh.read()).getroottree()
                buf = io.BytesIO()
                obj.write_c14n(buf)
                buf.seek(0, 0)
                content = buf.read()
            # check namespace definitions in root element
            expected = [
                b'xmlns="http://www.fdsn.org/xml/station/1"',
                b'xmlns:myns="http://test.myns.ns/"']
            for line in expected:
                assert line in content
            # check additional tags
            expected = [
                b'<myns:mynsNetworkTag>' +
                b'mynsNetworkTagValue' +
                b'</myns:mynsNetworkTag>',
                b'myns:mynsNetworkAttrib="mynsNetworkAttribValue"',
                b'<myns:mynsStationTag>' +
                b'mynsStationTagValue' +
                b'</myns:mynsStationTag>',
                b'myns:mynsStationAttrib="mynsStationAttribValue"',
                b'<myns:mynsChannelTag>' +
                b'mynsChannelTagValue' +
                b'</myns:mynsChannelTag>',
                b'myns:mynsChannelAttrib="mynsChannelAttribValue"',
                b'<myns:nested>',
                b'<myns:my_nested_tag1>' +
                b'12300000000.0' +
                b'</myns:my_nested_tag1>',
                b'<myns:my_nested_tag2 ' +
                b'myns:nestedAttribute1="nestedAttributeValue1">' +
                b'True' +
                b'</myns:my_nested_tag2>',
                b'</myns:nested>'
            ]
            for line in expected:
                assert line in content

    def test_write_with_extra_tags_and_read(self, testdata):
        """
        First tests that a StationXML file with additional
        custom "extra" tags gets written correctly. Then
        tests that when reading the written file again the
        extra tags are parsed correctly.
        """
        filename = testdata[
            "IRIS_single_channel_with_response_custom_tags.xml"]

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            inv = obspy.read_inventory(filename)
            assert len(w) == 0
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write file
            inv.write(tmpfile, format="STATIONXML")
            # check contents
            with open(tmpfile, "rb") as fh:
                # enforce reproducible attribute orders through write_c14n
                obj = etree.fromstring(fh.read()).getroottree()
                buf = io.BytesIO()
                obj.write_c14n(buf)
                buf.seek(0, 0)
                content = buf.read()
            # check namespace definitions in root element
            expected = [b'xmlns="http://www.fdsn.org/xml/station/1"',
                        b'xmlns:test="http://just.a.test/xmlns/1"'
                        ]
            for line in expected:
                assert line in content
            # check custom tags, nested custom tags, and attributes
            # at every level of the StationXML hierarchy
            expected = [
                # root
                b'test:customRootAttrib="testRootAttribute"',
                b'<test:CustomRootTag>testRootTag</test:CustomRootTag>',
                b'<test:CustomNestedRootTag>',
                b'<test:NestedTag1 nestedTagAttrib="testNestedAttribute">' +
                b'nestedRootTag1' +
                b'</test:NestedTag1>',
                b'<test:NestedTag2>nestedRootTag2</test:NestedTag2>',
                b'</test:CustomNestedRootTag>',
                # network
                b'test:customNetworkAttrib="testNetworkAttribute"',
                b'<test:CustomNetworkTag>' +
                b'testNetworkTag' +
                b'</test:CustomNetworkTag>',
                b'<test:CustomNestedNetworkTag>',
                b'<test:NestedTag1>nestedNetworkTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedNetworkTag2</test:NestedTag2>',
                b'</test:CustomNestedNetworkTag>',
                # station
                b'test:customStationAttrib="testStationAttribute"',
                b'<test:CustomStationTag>' +
                b'testStationTag' +
                b'</test:CustomStationTag>',
                b'<test:CustomNestedStationTag>',
                b'<test:NestedTag1>nestedStationTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedStationTag2</test:NestedTag2>',
                b'</test:CustomNestedStationTag>',
                # comment
                b'test:customCommentAttrib="testCommentAttribute"',
                b'<test:CustomCommentTag>' +
                b'testCommentTag' +
                b'</test:CustomCommentTag>',
                b'<test:CustomNestedCommentTag>',
                b'<test:NestedTag1>nestedCommentTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedCommentTag2</test:NestedTag2>',
                b'</test:CustomNestedCommentTag>',
                # person
                b'test:customPersonAttrib="testPersonAttribute"',
                b'<test:CustomPersonTag>testPersonTag</test:CustomPersonTag>',
                b'<test:CustomNestedPersonTag>',
                b'<test:NestedTag1>nestedPersonTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedPersonTag2</test:NestedTag2>',
                b'</test:CustomNestedPersonTag>',
                # phone
                b'test:customPhoneAttrib="testPhoneAttribute"',
                b'<test:CustomPhoneTag>testPhoneTag</test:CustomPhoneTag>',
                b'<test:CustomNestedPhoneTag>',
                b'<test:NestedTag1>nestedPhoneTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedPhoneTag2</test:NestedTag2>',
                b'</test:CustomNestedPhoneTag>',
                # site
                b'test:customSiteAttrib="testSiteAttribute"',
                b'<test:CustomSiteTag>testSiteTag</test:CustomSiteTag>',
                b'<test:CustomNestedSiteTag>',
                b'<test:NestedTag1>nestedSiteTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedSiteTag2</test:NestedTag2>',
                b'</test:CustomNestedSiteTag>',
                # equipment
                b'test:customEquipmentAttrib="testEquipmentAttribute"',
                b'<test:CustomEquipmentTag>' +
                b'testEquipmentTag' +
                b'</test:CustomEquipmentTag>',
                b'<test:CustomNestedEquipmentTag>',
                b'<test:NestedTag1>nestedEquipmentTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedEquipmentTag2</test:NestedTag2>',
                b'</test:CustomNestedEquipmentTag>',
                # operator
                b'test:customOperatorAttrib="testOperatorAttribute"',
                b'<test:CustomOperatorTag>' +
                b'testOperatorTag' +
                b'</test:CustomOperatorTag>',
                b'<test:CustomNestedOperatorTag>',
                b'<test:NestedTag1>nestedOperatorTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedOperatorTag2</test:NestedTag2>',
                b'</test:CustomNestedOperatorTag>',
                # external reference
                b'test:customExtRefAttrib="testExtRefAttribute"',
                b'<test:CustomExtRefTag>testExtRefTag</test:CustomExtRefTag>',
                b'<test:CustomNestedExtRefTag>',
                b'<test:NestedTag1>nestedExtRefTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedExtRefTag2</test:NestedTag2>',
                b'</test:CustomNestedExtRefTag>',
                # channel
                b'test:customChannelAttrib="testChannelAttribute"',
                b'<test:CustomChannelTag>' +
                b'testChannelTag' +
                b'</test:CustomChannelTag>',
                b'<test:CustomNestedChannelTag>',
                b'<test:NestedTag1>nestedChannelTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedChannelTag2</test:NestedTag2>',
                b'</test:CustomNestedChannelTag>',
                # response
                b'test:customResponseAttrib="testResponseAttribute"',
                b'<test:CustomResponseTag>' +
                b'testResponseTag' +
                b'</test:CustomResponseTag>',
                b'<test:CustomNestedResponseTag>',
                b'<test:NestedTag1>nestedResponseTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedResponseTag2</test:NestedTag2>',
                b'</test:CustomNestedResponseTag>',
                # data availability
                b'test:customDAAttrib="testDAAttribute"',
                b'<test:CustomDATag>testDATag</test:CustomDATag>',
                b'<test:CustomNestedDATag>',
                b'<test:NestedTag1>nestedDATag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedDATag2</test:NestedTag2>',
                b'</test:CustomNestedDATag>',
                # response stage (PolesZeros response stage)
                b'test:customStagePZAttrib="testStagePZAttribute"',
                b'<test:CustomStagePZTag>' +
                b'testStagePZTag' +
                b'</test:CustomStagePZTag>',
                b'<test:CustomNestedStagePZTag>',
                b'<test:NestedTag1>nestedStagePZTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedStagePZTag2</test:NestedTag2>',
                b'</test:CustomNestedStagePZTag>',
                # response stage (Coefficients response stage)
                b'test:customStageCoefAttrib="testStageCoefAttribute"',
                b'<test:CustomStageCoefTag>' +
                b'testStageCoefTag' +
                b'</test:CustomStageCoefTag>',
                b'<test:CustomNestedStageCoefTag>',
                b'<test:NestedTag1>nestedStageCoefTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedStageCoefTag2</test:NestedTag2>',
                b'</test:CustomNestedStageCoefTag>',
                # instrument sensitivity
                b'test:customSensitivityAttrib="testSensitivityAttribute"',
                b'<test:CustomSensitivityTag>' +
                b'testSensitivityTag' +
                b'</test:CustomSensitivityTag>',
                b'<test:CustomNestedSensitivityTag>',
                b'<test:NestedTag1>nestedSensitivityTag1</test:NestedTag1>',
                b'<test:NestedTag2>nestedSensitivityTag2</test:NestedTag2>',
                b'</test:CustomNestedSensitivityTag>'
            ]
            for line in expected:
                assert line in content
            # now, read again to test if it's parsed correctly..
            inv = obspy.read_inventory(tmpfile)

    def test_reading_file_with_empty_channel_object(self, testdata):
        """
        Tests reading a file with an empty channel object. This is strictly
        speaking not valid but we are forgiving.
        """
        filename = testdata["empty_channel.xml"]
        inv = obspy.read_inventory(filename)
        assert inv.get_contents() == \
            {'networks': ['IV'], 'stations': ['IV.LATE (Latera)'],
             'channels': []}

    def test_reading_channel_without_coordinates(self, testdata):
        """
        Tests reading a file with an empty channel object. This is strictly
        speaking not valid but we are forgiving.
        """
        filename = testdata["channel_without_coordinates.xml"]
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            inv = obspy.read_inventory(filename)

        # Should raise a warning that it could not read the channel without
        # coordinates.
        assert len(w) == 1
        assert w[0].message.args[0] == \
            ("Channel 00.BHZ of station LATE does not have a complete set of "
             "coordinates (latitude, longitude), elevation and depth and thus "
             "it cannot be read. It will not be part of "
             "the final inventory object.")

        assert inv.get_contents() == \
            {'networks': ['IV'], 'stations': ['IV.LATE (Latera)'],
             'channels': []}

    def test_units_during_identity_stage(self):
        """
        """
        t = obspy.UTCDateTime(2017, 1, 1)
        inv = obspy.read_inventory().select(station="RJOB", channel="EHZ",
                                            time=t)
        response = inv.get_response("BW.RJOB..EHZ", t)
        response.response_stages[0].input_units_description = "M/S"
        response.response_stages[0].output_units_description = "Volts"
        rstage_2 = ResponseStage(2, 1, 1, "V", "V",
                                 input_units_description="Volts",
                                 output_units_description="Volts")
        rstage_3 = ResponseStage(3, 1, 1, "V", "V",
                                 input_units_description="Volts",
                                 output_units_description="Volts")
        response.response_stages.insert(1, rstage_2)
        response.response_stages.insert(2, rstage_3)
        for i, rstage in enumerate(response.response_stages[3:]):
            rstage.stage_sequence_number = i + 4

        with io.BytesIO() as buf:
            inv.write(buf, format="stationxml", validate=True)
            buf.seek(0, 0)
            inv_2 = obspy.read_inventory(buf)

        response_2 = inv_2.get_response("BW.RJOB..EHZ", t)

        assert response == response_2
        assert response_2.response_stages[1].input_units == "V"
        assert response_2.response_stages[1].output_units == "V"
        assert response_2.response_stages[1].input_units_description == "Volts"
        assert response_2.response_stages[1].output_units_description == \
            "Volts"
        assert response_2.response_stages[2].input_units == "V"
        assert response_2.response_stages[2].output_units == "V"
        assert response_2.response_stages[2].input_units_description == "Volts"
        assert response_2.response_stages[2].output_units_description == \
            "Volts"

        # Also try from the other side.
        inv = obspy.read_inventory().select(station="RJOB", channel="EHZ",
                                            time=t)
        response = inv.get_response("BW.RJOB..EHZ", t)
        response.response_stages[0].input_units = None
        response.response_stages[0].output_units = None
        response.response_stages[0].input_units_description = None
        response.response_stages[0].output_units_description = None
        rstage_2 = ResponseStage(2, 1, 1, "V", "V",
                                 input_units_description="Volts",
                                 output_units_description="Volts")
        rstage_3 = ResponseStage(3, 1, 1, "V", "V",
                                 input_units_description="Volts",
                                 output_units_description="Volts")
        response.response_stages.insert(1, rstage_2)
        response.response_stages.insert(2, rstage_3)
        for i, rstage in enumerate(response.response_stages[3:]):
            rstage.stage_sequence_number = i + 4

        with io.BytesIO() as buf:
            inv.write(buf, format="stationxml")
            buf.seek(0, 0)
            inv_2 = obspy.read_inventory(buf)

        response_2 = inv_2.get_response("BW.RJOB..EHZ", t)
        # Set these to None manually as the autocorrection during parsing will
        # set it.
        response_2.response_stages[0].input_units = None
        response_2.response_stages[0].input_units_description = None

        assert response == response_2
        assert response_2.response_stages[1].input_units == "V"
        assert response_2.response_stages[1].output_units == "V"
        assert response_2.response_stages[1].input_units_description == "Volts"
        assert response_2.response_stages[1].output_units_description == \
            "Volts"
        assert response_2.response_stages[2].input_units == "V"
        assert response_2.response_stages[2].output_units == "V"
        assert response_2.response_stages[2].input_units_description == "Volts"
        assert response_2.response_stages[2].output_units_description == \
            "Volts"

    def test_reading_full_stationxml_1_0_file(self, testdata):
        """
        Tests reading a fully filled StationXML 1.0 file.
        """
        filename = testdata["full_random_stationxml_1_0.xml"]
        inv = obspy.read_inventory(filename, format='STATIONXML')
        lats = [cha.latitude for net in inv for sta in net for cha in sta]
        # for now just check that all expected channels are there.. test could
        # be much improved
        assert lats == [
            -53.12, 44.77, 63.39, 12.46, -13.16, -84.44, 43.9, -88.41]

    def test_read_with_level(self, testdata):
        """
        Tests reading StationXML with specifying the level of detail.
        """
        path = testdata['stationxml_BK.CMB.__.LKS.xml']
        inv_stationxml_no_level = _read_stationxml(path)
        inv_stationxml_response = _read_stationxml(path, level='response')
        inv_stationxml_channel = _read_stationxml(path, level='channel')
        inv_stationxml_station = _read_stationxml(path, level='station')
        inv_stationxml_network = _read_stationxml(path, level='network')
        # test reading through plugin
        assert obspy.read_inventory(
            path, format='STATIONXML', level='station') == \
            inv_stationxml_station
        # test reading default which should be equivalent to reading response
        # level
        assert inv_stationxml_no_level == inv_stationxml_response
        # test reading response level
        assert len(inv_stationxml_response) == 1
        assert len(inv_stationxml_response[0]) == 1
        assert len(inv_stationxml_response[0][0]) == 1
        assert inv_stationxml_response[0][0][0].response is not None
        # test reading channel level
        assert len(inv_stationxml_channel) == 1
        assert len(inv_stationxml_channel[0]) == 1
        assert len(inv_stationxml_channel[0][0]) == 1
        assert inv_stationxml_channel[0][0][0].response is None
        # test reading station level
        assert len(inv_stationxml_station) == 1
        assert len(inv_stationxml_station[0]) == 1
        assert len(inv_stationxml_station[0][0]) == 0
        # test reading station level
        assert len(inv_stationxml_network) == 1
        assert len(inv_stationxml_network[0]) == 0

    def test_read_basic_responsestage_with_decimation(self, testdata):
        """
        Make sure basic ResponseStage elements that have decimation information
        do not lose that information.
        """
        path = testdata['F1_423_small.xml']
        inv = _read_stationxml(path)
        stage = inv[0][0][0].response.response_stages[0]
        assert stage.decimation_correction == 0.4
        assert stage.decimation_delay == 0.5
        assert stage.decimation_factor == 2
        assert stage.decimation_input_sample_rate == 1024000.0
        assert stage.decimation_offset == 1

    def test_do_not_write_None_IO_units(self):
        """
        Regression test for #3572

        Makes sure that in InputUnits and OutputUnits if the optional unit
        description is None we do not write a tag with text value "None"
        """
        resp = Response.from_paz(
            [], [], 2000, input_units='m/s', output_units='count')
        channels = [Channel('LHZ', '00', 0, 0, 0, 0, response=resp, dip=-90)]
        stations = [Station('STA', 0, 0, 0, channels=channels)]
        networks = [Network('XX', stations=stations)]
        inv = Inventory(networks=networks)
        with io.BytesIO() as bio:
            inv.write(bio, format='STATIONXML')
            bio.seek(0)
            xml = etree.parse(bio)
        namespaces = {'ns': 'http://www.fdsn.org/xml/station/1'}
        # xpath will return an empty list if the given tag is missing or has a
        # NULL value (i.e.empty tag "<Description/>")
        matches = xml.xpath(
            '/ns:FDSNStationXML/ns:Network/ns:Station/ns:Channel/ns:Response/'
            'ns:InstrumentSensitivity/ns:InputUnits/ns:Description',
            namespaces=namespaces)
        # element should not be present, i.e. matches an empty list.
        assert not matches
        matches = xml.xpath(
            '/ns:FDSNStationXML/ns:Network/ns:Station/ns:Channel/ns:Response/'
            'ns:InstrumentSensitivity/ns:OutputUnits/ns:Description',
            namespaces=namespaces)
        # element should not be present, i.e. matches an empty list.
        assert not matches
        # if it is present, at least it should have empty value, but let's just
        # make sure it's not present at all for now
        # assert matches[0].text is None



# -*- coding: utf-8 -*-
import io
import math
import warnings

import pytest
from lxml import etree

from obspy.core.event import (Catalog, Event, FocalMechanism, Magnitude,
                              MomentTensor, Origin, Pick, ResourceIdentifier,
                              Tensor, WaveformStreamID, read_events,
                              EventDescription)
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.util import AttribDict
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.util.testing import compare_xml_strings
from obspy.io.quakeml.core import Pickler, _read_quakeml, _write_quakeml


# lxml < 2.3 seems not to ship with RelaxNG schema parser and namespace support
IS_RECENT_LXML = False
version = float(etree.__version__.rsplit('.', 1)[0])
if version >= 2.3:
    IS_RECENT_LXML = True


def assert_no_extras(obj, verbose=False):
    """
    Helper routine to make sure no information ends up in 'extra' when there is
    not supposed to be anything in there. (to make sure, after changes in
    #2466)
    """
    assert getattr(obj, 'extra', None) is None
    if verbose:
        print('no extras in {!s}'.format(obj))
    if isinstance(obj, Catalog):
        for event in obj:
            assert_no_extras(event, verbose=verbose)
        return
    # recurse deeper if an event-type object
    for name, _type in getattr(obj, '_properties', []):
        assert_no_extras(getattr(obj, name), verbose=verbose)


class TestQuakeML():
    """
    Test suite for obspy.io.quakeml
    """
    def test_read_quakeml(self, testdata):
        """
        """
        # IRIS
        filename = testdata['iris_events.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 2
        assert catalog[0].resource_id == \
            ResourceIdentifier(
                'smi:www.iris.edu/ws/event/query?eventId=3279407')
        assert catalog[1].resource_id == \
            ResourceIdentifier(
                'smi:www.iris.edu/ws/event/query?eventId=2318174')
        # NERIES
        catalog = _read_quakeml(testdata['neries_events.xml'])
        assert_no_extras(catalog)
        assert len(catalog) == 3
        assert catalog[0].resource_id == \
            ResourceIdentifier('quakeml:eu.emsc/event/20120404_0000041')
        assert catalog[1].resource_id == \
            ResourceIdentifier('quakeml:eu.emsc/event/20120404_0000038')
        assert catalog[2].resource_id == \
            ResourceIdentifier('quakeml:eu.emsc/event/20120404_0000039')

    def test_usgs_eventype(self, testdata):
        filename = testdata['usgs_event.xml']
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("ignore")
            catalog = _read_quakeml(filename)
        # there are some custom namespace attributes
        assert len(catalog[0].extra) == 3
        assert len(catalog[0].origins[0].extra) == 4
        assert len(catalog) == 1
        assert catalog[0].event_type == 'quarry blast'

    def test_event(self, testdata):
        """
        Tests Event object.
        """
        filename = testdata['quakeml_1.2_event.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        event = catalog[0]
        assert event.resource_id == \
            ResourceIdentifier('smi:ch.ethz.sed/event/historical/1165')
        # enums
        assert event.event_type == 'earthquake'
        assert event.event_type_certainty == 'suspected'
        # comments
        assert len(event.comments) == 2
        c = event.comments
        assert c[0].text == 'Relocated after re-evaluation'
        assert c[0].resource_id is None
        assert c[0].creation_info.agency_id == 'EMSC'
        assert c[1].text == 'Another comment'
        assert c[1].resource_id == \
            ResourceIdentifier(id="smi:some/comment/id/number_3")
        assert c[1].creation_info is None
        # event descriptions
        assert len(event.event_descriptions) == 3
        d = event.event_descriptions
        assert d[0].text == '1906 San Francisco Earthquake'
        assert d[0].type == 'earthquake name'
        assert d[1].text == 'NEAR EAST COAST OF HONSHU, JAPAN'
        assert d[1].type == 'Flinn-Engdahl region'
        assert d[2].text == 'free-form string'
        assert d[2].type is None
        # creation info
        assert event.creation_info.author == "Erika Mustermann"
        assert event.creation_info.agency_id == "EMSC"
        assert event.creation_info.author_uri == \
            ResourceIdentifier("smi:smi-registry/organization/EMSC")
        assert event.creation_info.agency_uri == \
            ResourceIdentifier("smi:smi-registry/organization/EMSC")
        assert event.creation_info.creation_time == \
            UTCDateTime("2012-04-04T16:40:50+00:00")
        assert event.creation_info.version == "1.0.1"
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_origin(self, testdata):
        """
        Tests Origin object.
        """
        filename = testdata['quakeml_1.2_origin.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].origins) == 1
        origin = catalog[0].origins[0]
        assert origin.resource_id == \
            ResourceIdentifier(
                'smi:www.iris.edu/ws/event/query?originId=7680412')
        assert origin.time == UTCDateTime("2011-03-11T05:46:24.1200")
        assert origin.latitude == 38.297
        assert origin.latitude_errors.lower_uncertainty is None
        assert origin.longitude == 142.373
        assert origin.longitude_errors.uncertainty is None
        assert origin.depth == 29.0
        assert origin.depth_errors.confidence_level == 50.0
        assert origin.depth_type == "from location"
        assert origin.method_id == \
            ResourceIdentifier(id="smi:some/method/NA")
        assert origin.time_fixed is None
        assert not origin.epicenter_fixed
        assert origin.reference_system_id == \
            ResourceIdentifier(id="smi:some/reference/muh")
        assert origin.earth_model_id == \
            ResourceIdentifier(id="smi:same/model/maeh")
        assert origin.region == 'Kalamazoo'
        assert origin.evaluation_mode == "manual"
        assert origin.evaluation_status == "preliminary"
        assert origin.origin_type == "hypocenter"
        # composite times
        assert len(origin.composite_times) == 2
        c = origin.composite_times
        assert c[0].year == 2029
        assert c[0].month is None
        assert c[0].day is None
        assert c[0].hour == 12
        assert c[0].minute is None
        assert c[0].second is None
        assert c[1].year is None
        assert c[1].month is None
        assert c[1].day is None
        assert c[1].hour == 1
        assert c[1].minute is None
        assert c[1].second == 29.124234
        # quality
        assert origin.quality.used_station_count == 16
        assert origin.quality.standard_error == 0
        assert origin.quality.azimuthal_gap == 231
        assert origin.quality.maximum_distance == 53.03
        assert origin.quality.minimum_distance == 2.45
        assert origin.quality.associated_phase_count is None
        assert origin.quality.associated_station_count is None
        assert origin.quality.depth_phase_count is None
        assert origin.quality.secondary_azimuthal_gap is None
        assert origin.quality.ground_truth_level is None
        assert origin.quality.median_distance is None
        # comments
        assert len(origin.comments) == 2
        c = origin.comments
        assert c[0].text == 'Some comment'
        assert c[0].resource_id == \
            ResourceIdentifier(id="smi:some/comment/reference")
        assert c[0].creation_info.author == 'EMSC'
        assert c[1].resource_id is None
        assert c[1].creation_info is None
        assert c[1].text == 'Another comment'
        # creation info
        assert origin.creation_info.author == "NEIC"
        assert origin.creation_info.agency_id is None
        assert origin.creation_info.author_uri is None
        assert origin.creation_info.agency_uri is None
        assert origin.creation_info.creation_time is None
        assert origin.creation_info.version is None
        # origin uncertainty
        u = origin.origin_uncertainty
        assert u.preferred_description == "uncertainty ellipse"
        assert u.horizontal_uncertainty == 9000
        assert u.min_horizontal_uncertainty == 6000
        assert u.max_horizontal_uncertainty == 10000
        assert u.azimuth_max_horizontal_uncertainty == 80.0
        # confidence ellipsoid
        c = u.confidence_ellipsoid
        assert c.semi_intermediate_axis_length == 2123.0
        assert c.major_axis_rotation == 5.123
        assert c.major_axis_plunge == 3.123
        assert c.semi_minor_axis_length == 1123.0
        assert c.semi_major_axis_length == 123.0
        assert c.major_axis_azimuth == 4.123
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_magnitude(self, testdata):
        """
        Tests Magnitude object.
        """
        filename = testdata['quakeml_1.2_magnitude.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].magnitudes) == 1
        mag = catalog[0].magnitudes[0]
        assert mag.resource_id == \
            ResourceIdentifier('smi:ch.ethz.sed/magnitude/37465')
        assert mag.mag == 5.5
        assert mag.mag_errors.uncertainty == 0.1
        assert mag.magnitude_type == 'MS'
        assert mag.method_id == \
            ResourceIdentifier(
                'smi:ch.ethz.sed/magnitude/generic/surface_wave_magnitude')
        assert mag.station_count == 8
        assert mag.evaluation_status == 'preliminary'
        # comments
        assert len(mag.comments) == 2
        c = mag.comments
        assert c[0].text == 'Some comment'
        assert c[0].resource_id == \
            ResourceIdentifier(id="smi:some/comment/id/muh")
        assert c[0].creation_info.author == 'EMSC'
        assert c[1].creation_info is None
        assert c[1].text == 'Another comment'
        assert c[1].resource_id is None
        # creation info
        assert mag.creation_info.author == "NEIC"
        assert mag.creation_info.agency_id is None
        assert mag.creation_info.author_uri is None
        assert mag.creation_info.agency_uri is None
        assert mag.creation_info.creation_time is None
        assert mag.creation_info.version is None
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_station_magnitude_contribution(self, testdata):
        """
        Tests the station magnitude contribution object.
        """
        filename = testdata['quakeml_1.2_stationmagnitudecontributions.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].magnitudes) == 1
        assert len(
            catalog[0].magnitudes[0].station_magnitude_contributions) == 2
        # Check the first stationMagnitudeContribution object.
        stat_contrib = \
            catalog[0].magnitudes[0].station_magnitude_contributions[0]
        assert stat_contrib.station_magnitude_id.id == \
            "smi:ch.ethz.sed/magnitude/station/881342"
        assert stat_contrib.weight == 0.77
        assert stat_contrib.residual == 0.02
        # Check the second stationMagnitudeContribution object.
        stat_contrib = \
            catalog[0].magnitudes[0].station_magnitude_contributions[1]
        assert stat_contrib.station_magnitude_id.id == \
            "smi:ch.ethz.sed/magnitude/station/881334"
        assert stat_contrib.weight == 0.
        assert stat_contrib.residual == 0.

        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_station_magnitude(self, testdata):
        """
        Tests StationMagnitude object.
        """
        filename = testdata['quakeml_1.2_stationmagnitude.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].station_magnitudes) == 1
        mag = catalog[0].station_magnitudes[0]
        # Assert the actual StationMagnitude object. Everything that is not set
        # in the QuakeML file should be set to None.
        assert mag.resource_id == \
            ResourceIdentifier("smi:ch.ethz.sed/magnitude/station/881342")
        assert mag.origin_id == \
            ResourceIdentifier('smi:some/example/id')
        assert mag.mag == 6.5
        assert mag.mag_errors.uncertainty == 0.2
        assert mag.station_magnitude_type == 'MS'
        assert mag.amplitude_id == \
            ResourceIdentifier("smi:ch.ethz.sed/amplitude/824315")
        assert mag.method_id == \
            ResourceIdentifier(
                "smi:ch.ethz.sed/magnitude/generic/surface_wave_magnitude")
        assert mag.waveform_id == \
            WaveformStreamID(network_code='BW', station_code='FUR',
                             resource_uri="smi:ch.ethz.sed/waveform/201754")
        assert mag.creation_info is None
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_data_used_in_moment_tensor(self, testdata):
        """
        Tests the data used objects in moment tensors.
        """
        filename = testdata['quakeml_1.2_data_used.xml']

        # Test reading first.
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        event = catalog[0]

        assert len(event.focal_mechanisms), 2
        # First focmec contains only one data used element.
        assert len(event.focal_mechanisms[0].moment_tensor.data_used) == 1
        du = event.focal_mechanisms[0].moment_tensor.data_used[0]
        assert du.wave_type == "body waves"
        assert du.station_count == 88
        assert du.component_count == 166
        assert du.shortest_period == 40.0
        # Second contains three. focmec contains only one data used element.
        assert len(event.focal_mechanisms[1].moment_tensor.data_used) == 3
        du = event.focal_mechanisms[1].moment_tensor.data_used
        assert du[0].wave_type == "body waves"
        assert du[0].station_count == 88
        assert du[0].component_count == 166
        assert du[0].shortest_period == 40.0
        assert du[1].wave_type == "surface waves"
        assert du[1].station_count == 96
        assert du[1].component_count == 189
        assert du[1].shortest_period == 50.0
        assert du[2].wave_type == "mantle waves"
        assert du[2].station_count == 41
        assert du[2].component_count == 52
        assert du[2].shortest_period == 125.0

        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_arrival(self, testdata):
        """
        Tests Arrival object.
        """
        filename = testdata['quakeml_1.2_arrival.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].origins[0].arrivals) == 2
        ar = catalog[0].origins[0].arrivals[0]
        # Test the actual Arrival object. Everything not set in the QuakeML
        # file should be None.
        assert ar.pick_id == \
            ResourceIdentifier('smi:ch.ethz.sed/pick/117634')
        assert ar.phase == 'Pn'
        assert ar.azimuth == 12.0
        assert ar.distance == 0.5
        assert ar.takeoff_angle == 11.0
        assert ar.takeoff_angle_errors.uncertainty == 0.2
        assert ar.time_residual == 1.6
        assert ar.horizontal_slowness_residual == 1.7
        assert ar.backazimuth_residual == 1.8
        assert ar.time_weight == 0.48
        assert ar.horizontal_slowness_weight == 0.49
        assert ar.backazimuth_weight == 0.5
        assert ar.earth_model_id == \
            ResourceIdentifier('smi:ch.ethz.sed/earthmodel/U21')
        assert len(ar.comments) == 1
        assert ar.creation_info.author == "Erika Mustermann"
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_pick(self, testdata):
        """
        Tests Pick object.
        """
        filename = testdata['quakeml_1.2_pick.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].picks) == 2
        pick = catalog[0].picks[0]
        assert pick.resource_id == \
            ResourceIdentifier('smi:ch.ethz.sed/pick/117634')
        assert pick.time == UTCDateTime('2005-09-18T22:04:35Z')
        assert pick.time_errors.uncertainty == 0.012
        assert pick.waveform_id == \
            WaveformStreamID(network_code='BW', station_code='FUR',
                             resource_uri='smi:ch.ethz.sed/waveform/201754')
        assert pick.filter_id == \
            ResourceIdentifier('smi:ch.ethz.sed/filter/lowpass/standard')
        assert pick.method_id == \
            ResourceIdentifier('smi:ch.ethz.sed/picker/autopicker/6.0.2')
        assert pick.backazimuth == 44.0
        assert pick.onset == 'impulsive'
        assert pick.phase_hint == 'Pn'
        assert pick.polarity == 'positive'
        assert pick.evaluation_mode == "manual"
        assert pick.evaluation_status == "confirmed"
        assert len(pick.comments) == 2
        assert pick.creation_info.author == "Erika Mustermann"
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_focalmechanism(self, testdata):
        """
        Tests FocalMechanism object.
        """
        filename = testdata['quakeml_1.2_focalmechanism.xml']
        catalog = _read_quakeml(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        assert len(catalog[0].focal_mechanisms) == 2
        fm = catalog[0].focal_mechanisms[0]
        # general
        assert fm.resource_id == \
            ResourceIdentifier('smi:ISC/fmid=292309')
        assert len(fm.waveform_id) == 2
        assert fm.waveform_id[0].network_code == 'BW'
        assert fm.waveform_id[0].station_code == 'FUR'
        assert fm.waveform_id[0].resource_uri == \
            ResourceIdentifier(id="smi:ch.ethz.sed/waveform/201754")
        assert isinstance(fm.waveform_id[0], WaveformStreamID)
        assert fm.triggering_origin_id == \
            ResourceIdentifier('smi:local/originId=7680412')
        assert round(abs(fm.azimuthal_gap-0.123), 7) == 0
        assert fm.station_polarity_count == 987
        assert round(abs(fm.misfit-1.234), 7) == 0
        assert round(abs(fm.station_distribution_ratio-2.345), 7) == 0
        assert fm.method_id == \
            ResourceIdentifier('smi:ISC/methodID=Best_double_couple')
        # comments
        assert len(fm.comments) == 2
        c = fm.comments
        assert c[0].text == 'Relocated after re-evaluation'
        assert c[0].resource_id is None
        assert c[0].creation_info.agency_id == 'MUH'
        assert c[1].text == 'Another MUH'
        assert c[1].resource_id == \
            ResourceIdentifier(id="smi:some/comment/id/number_3")
        assert c[1].creation_info is None
        # creation info
        assert fm.creation_info.author == "Erika Mustermann"
        assert fm.creation_info.agency_id == "MUH"
        assert fm.creation_info.author_uri == \
            ResourceIdentifier("smi:smi-registry/organization/MUH")
        assert fm.creation_info.agency_uri == \
            ResourceIdentifier("smi:smi-registry/organization/MUH")
        assert fm.creation_info.creation_time == \
            UTCDateTime("2012-04-04T16:40:50+00:00")
        assert fm.creation_info.version == "1.0.1"
        # nodalPlanes
        assert round(abs(fm.nodal_planes.nodal_plane_1.strike-346.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_1.dip-57.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_1.rake-75.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_2.strike-193.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_2.dip-36.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_2.rake-112.0), 7) == 0
        assert fm.nodal_planes.preferred_plane == 2
        # principalAxes
        assert round(abs(fm.principal_axes.t_axis.azimuth-216.0), 7) == 0
        assert round(abs(fm.principal_axes.t_axis.plunge-73.0), 7) == 0
        assert round(abs(fm.principal_axes.t_axis.length-1.050e+18), 7) == 0
        assert round(abs(fm.principal_axes.p_axis.azimuth-86.0), 7) == 0
        assert round(abs(fm.principal_axes.p_axis.plunge-10.0), 7) == 0
        assert round(abs(fm.principal_axes.p_axis.length--1.180e+18), 7) == 0
        assert fm.principal_axes.n_axis.azimuth is None
        assert fm.principal_axes.n_axis.plunge is None
        assert fm.principal_axes.n_axis.length is None
        # momentTensor
        mt = fm.moment_tensor
        assert mt.resource_id == \
            ResourceIdentifier('smi:ISC/mtid=123321')
        assert mt.derived_origin_id == \
            ResourceIdentifier('smi:ISC/origid=13145006')
        assert round(abs(mt.scalar_moment-1.100e+18), 7) == 0
        assert round(abs(mt.tensor.m_rr-9.300e+17), 7) == 0
        assert round(abs(mt.tensor.m_tt-1.700e+17), 7) == 0
        assert round(abs(mt.tensor.m_pp--1.100e+18), 7) == 0
        assert round(abs(mt.tensor.m_rt--2.200e+17), 7) == 0
        assert round(abs(mt.tensor.m_rp-4.000e+17), 7) == 0
        assert round(abs(mt.tensor.m_tp-3.000e+16), 7) == 0
        assert round(abs(mt.clvd-0.22), 7) == 0
        # exporting back to XML should result in the same document
        with open(filename, "rb") as fp:
            original = fp.read()
        processed = Pickler().dumps(catalog)
        compare_xml_strings(original, processed)

    def test_write_quakeml(self, testdata):
        """
        Tests writing a QuakeML document.
        """
        filename = testdata['qml-example-1.2-RC3.xml']
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            catalog = _read_quakeml(filename)
            assert_no_extras(catalog)
            assert len(catalog), 1
            _write_quakeml(catalog, tmpfile, validate=IS_RECENT_LXML)
            # Read file again. Avoid the (legit) warning about the already used
            # resource identifiers.
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("ignore")
                catalog2 = _read_quakeml(tmpfile)
                assert_no_extras(catalog2)
        assert len(catalog2), 1

    def test_read_events(self, testdata):
        """
        Tests reading a QuakeML document via read_events.
        """
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            catalog = read_events(testdata['neries_events.xml'])
            assert_no_extras(catalog)
            assert len(catalog), 3
            catalog.write(tmpfile, format='QUAKEML')
            # Read file again. Avoid the (legit) warning about the already used
            # resource identifiers.
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("ignore")
                catalog2 = read_events(tmpfile)
                assert_no_extras(catalog2)
        assert len(catalog2), 3

    @pytest.mark.skipif(not IS_RECENT_LXML, reason="lxml >= 2.3 is required")
    def test_enums(self, datapath):
        """
        Parses the QuakeML xsd scheme definition and checks if all enums are
        correctly defined.

        This is a very strict test against the xsd scheme file of QuakeML 1.2.
        If obspy.core.event will ever be more loosely coupled to QuakeML this
        test WILL HAVE to be changed.
        """
        from lxml.etree import parse

        xsd_enum_definitions = {}
        xsd_file = datapath.parent.parent / "data" / "QuakeML-BED-1.2.xsd"
        root = parse(xsd_file).getroot()

        # Get all enums from the xsd file.
        nsmap = {k: v for k, v in root.nsmap.items() if k is not None}
        for stype in root.findall("xs:simpleType", namespaces=nsmap):
            type_name = stype.get("name")
            restriction = stype.find("xs:restriction", namespaces=nsmap)
            if restriction is None:
                continue
            if restriction.get("base") != "xs:string":
                continue
            enums = restriction.findall(
                "xs:enumeration", namespaces=nsmap)
            if not enums:
                continue
            enums = [_i.get("value") for _i in enums]
            xsd_enum_definitions[type_name] = enums

        # Now import all enums and check if they are correct.
        from obspy.core.event import header as event_header
        from obspy.core.util import Enum
        all_enums = {}
        for module_item_name in dir(event_header):
            module_item = getattr(event_header, module_item_name)
            if not isinstance(module_item, Enum):
                continue
            # Assign clearer names.
            enum_name = module_item_name
            enum_values = [_i.lower() for _i in module_item.keys()]
            all_enums[enum_name] = enum_values
        # Now loop over all enums defined in the xsd file and check them.
        for enum_name, enum_items in xsd_enum_definitions.items():
            assert enum_name in all_enums.keys()
            # Check that also all enum items are available.
            all_items = all_enums[enum_name]
            all_items = [_i.lower() for _i in all_items]
            for enum_item in enum_items:
                if enum_item.lower() not in all_items:  # pragma: no cover
                    msg = "Value '%s' not in Enum '%s'" % \
                        (enum_item, enum_name)
                    raise Exception(msg)
            # Check if there are too many items.
            if len(all_items) != len(enum_items):  # pragma: no cover
                additional_items = [_i for _i in all_items
                                    if _i.lower() not in enum_items]
                msg = "Enum {enum_name} has the following additional items" + \
                    " not defined in the xsd style sheet:\n\t{enumerations}"
                msg = msg.format(enum_name=enum_name,
                                 enumerations=", ".join(additional_items))
                raise Exception(msg)

    def test_read_string(self, testdata):
        """
        Test reading a QuakeML string/unicode object via read_events.
        """
        with open(testdata['neries_events.xml'], 'rb') as fp:
            data = fp.read()

        catalog = read_events(data)
        assert_no_extras(catalog)
        assert len(catalog) == 3

    def test_preferred_tags(self, testdata):
        """
        Testing preferred magnitude, origin and focal mechanism tags
        """
        # testing empty event
        ev = Event()
        assert ev.preferred_origin() is None
        assert ev.preferred_magnitude() is None
        assert ev.preferred_focal_mechanism() is None
        # testing existing event
        filename = testdata['preferred.xml']
        catalog = read_events(filename)
        assert_no_extras(catalog)
        assert len(catalog) == 1
        ev_str = "Event:\t2012-12-12T05:46:24.120000Z | +38.297, +142.373 " + \
                 "| 2.0  MW"
        assert ev_str in str(catalog.events[0])
        # testing ids
        ev = catalog.events[0]
        assert 'smi:orig2' == ev.preferred_origin_id
        assert 'smi:mag2' == ev.preferred_magnitude_id
        assert 'smi:fm2' == ev.preferred_focal_mechanism_id
        # testing objects
        assert ev.preferred_origin() == ev.origins[1]
        assert ev.preferred_magnitude() == ev.magnitudes[1]
        assert ev.preferred_focal_mechanism() == ev.focal_mechanisms[1]

    def test_creating_minimal_quakeml_with_mt(self):
        """
        Tests the creation of a minimal QuakeML containing origin, magnitude
        and moment tensor.
        """
        # Rotate into physical domain
        lat, lon, depth, org_time = 10.0, -20.0, 12000, UTCDateTime(2012, 1, 1)
        mrr, mtt, mpp, mtr, mpr, mtp = 1E18, 2E18, 3E18, 3E18, 2E18, 1E18
        scalar_moment = math.sqrt(
            mrr ** 2 + mtt ** 2 + mpp ** 2 + mtr ** 2 + mpr ** 2 + mtp ** 2)
        moment_magnitude = 0.667 * (math.log10(scalar_moment) - 9.1)

        # Initialise event
        ev = Event(event_type="earthquake")

        ev_origin = Origin(time=org_time, latitude=lat, longitude=lon,
                           depth=depth, resource_id=ResourceIdentifier())
        ev.origins.append(ev_origin)

        # populate event moment tensor
        ev_tensor = Tensor(m_rr=mrr, m_tt=mtt, m_pp=mpp, m_rt=mtr, m_rp=mpr,
                           m_tp=mtp)

        ev_momenttensor = MomentTensor(tensor=ev_tensor)
        ev_momenttensor.scalar_moment = scalar_moment
        ev_momenttensor.derived_origin_id = ev_origin.resource_id

        ev_focalmechanism = FocalMechanism(moment_tensor=ev_momenttensor)
        ev.focal_mechanisms.append(ev_focalmechanism)

        # populate event magnitude
        ev_magnitude = Magnitude()
        ev_magnitude.mag = moment_magnitude
        ev_magnitude.magnitude_type = 'Mw'
        ev_magnitude.evaluation_mode = 'automatic'
        ev.magnitudes.append(ev_magnitude)

        # write QuakeML file
        cat = Catalog(events=[ev])
        memfile = io.BytesIO()
        cat.write(memfile, format="quakeml", validate=IS_RECENT_LXML)

        memfile.seek(0, 0)
        new_cat = _read_quakeml(memfile)
        assert_no_extras(new_cat)
        assert len(new_cat) == 1
        event = new_cat[0]
        assert len(event.origins) == 1
        assert len(event.magnitudes) == 1
        assert len(event.focal_mechanisms) == 1
        org = event.origins[0]
        mag = event.magnitudes[0]
        fm = event.focal_mechanisms[0]
        assert org.latitude == lat
        assert org.longitude == lon
        assert org.depth == depth
        assert org.time == org_time
        # Moment tensor.
        mt = fm.moment_tensor.tensor
        assert (fm.moment_tensor.scalar_moment - scalar_moment) / \
            scalar_moment < scalar_moment * 1E-10
        assert mt.m_rr == mrr
        assert mt.m_pp == mpp
        assert mt.m_tt == mtt
        assert mt.m_rt == mtr
        assert mt.m_rp == mpr
        assert mt.m_tp == mtp
        # Mag
        assert round(abs(mag.mag-moment_magnitude), 7) == 0
        assert mag.magnitude_type == "Mw"
        assert mag.evaluation_mode == "automatic"

    def test_read_equivalence(self, testdata):
        """
        See #662.
        Tests if _read_quakeml() and read_events() return the same results.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            cat1 = read_events(testdata['neries_events.xml'])
            cat2 = _read_quakeml(testdata['neries_events.xml'])
        assert cat1 == cat2

    def test_reading_twice_raises_no_warning(self, testdata):
        """
        Tests that reading a QuakeML file twice does not raise a warnings.

        Not an extensive test but likely good enough.
        """
        filename = testdata['qml-example-1.2-RC3.xml']

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat1 = _read_quakeml(filename)
            assert len(w) == 0
            cat2 = _read_quakeml(filename)
            assert len(w) == 0
        assert_no_extras(cat1)

        assert cat1 == cat2

    def test_read_amplitude_time_window(self, testdata):
        """
        Tests reading an QuakeML Amplitude with TimeWindow.
        """
        filename = testdata['qml-example-1.2-RC3.xml']

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat = _read_quakeml(filename)
            assert len(w) == 0

        assert len(cat[0].amplitudes) == 1
        amp = cat[0].amplitudes[0]
        assert amp.type == "A"
        assert amp.category == "point"
        assert amp.unit == "m/s"
        assert amp.generic_amplitude == 1e-08
        assert amp.time_window.begin == 0.0
        assert amp.time_window.end == 0.51424
        assert amp.time_window.reference == \
            UTCDateTime("2007-10-10T14:40:39.055")

    def test_write_amplitude_time_window(self, testdata):
        """
        Tests writing an QuakeML Amplitude with TimeWindow.
        """
        filename = testdata['qml-example-1.2-RC3.xml']

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat = _read_quakeml(filename)
            assert len(w) == 0

        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            cat.write(tmpfile, format='QUAKEML')
            with open(tmpfile, "rb") as fh:
                lines = fh.readlines()

            firstline = 45
            while b"<amplitude " not in lines[firstline]:
                firstline += 1

            got = [lines[i_].strip()
                   for i_ in range(firstline, firstline + 13)]
            expected = [
                b'<amplitude publicID="smi:nz.org.geonet/event/2806038g/'
                b'amplitude/1/modified">',
                b'<genericAmplitude>',
                b'<value>1e-08</value>',
                b'</genericAmplitude>',
                b'<type>A</type>',
                b'<category>point</category>',
                b'<unit>m/s</unit>',
                b'<timeWindow>',
                b'<reference>2007-10-10T14:40:39.055000Z</reference>',
                b'<begin>0.0</begin>',
                b'<end>0.51424</end>',
                b'</timeWindow>',
                b'</amplitude>']
            assert got == expected

    def test_write_with_extra_tags_and_read(self, testdata):
        """
        Tests that a QuakeML file with additional custom "extra" tags gets
        written correctly and that when reading it again the extra tags are
        parsed correctly.
        """
        filename = testdata['quakeml_1.2_origin.xml']

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat = _read_quakeml(filename)
            assert_no_extras(cat)
            assert len(w) == 0

        # add some custom tags to first event:
        #  - tag with explicit namespace but no explicit ns abbreviation
        #  - tag without explicit namespace (gets obspy default ns)
        #  - tag with explicit namespace and namespace abbreviation
        my_extra = AttribDict(
            {'public': {'value': False,
                        'namespace': 'http://some-page.de/xmlns/1.0',
                        'attrib': {'some_attrib': 'some_value',
                                   'another_attrib': 'another_value'}},
             'custom': {'value': 'True',
                        'namespace': 'http://test.org/xmlns/0.1'},
             'new_tag': {'value': 1234,
                         'namespace': 'http://test.org/xmlns/0.1'},
             'tX': {'value': UTCDateTime('2013-01-02T13:12:14.600000Z'),
                    'namespace': 'http://test.org/xmlns/0.1'},
             'dataid': {'namespace': 'http://anss.org/xmlns/catalog/0.1',
                        'type': 'attribute', 'value': '00999999'},
             # some nested tags :
             'quantity': {'namespace': 'http://some-page.de/xmlns/1.0',
                          'attrib': {'attrib1': 'attrib_value1',
                                     'attrib2': 'attrib_value2'},
                          'value': {
                              'my_nested_tag1': {
                                  'namespace': 'http://some-page.de/xmlns/1.0',
                                  'value': 1.23E10},
                              'my_nested_tag2': {
                                  'namespace': 'http://some-page.de/xmlns/1.0',
                                  'value': False}}}})
        nsmap = {'ns0': 'http://test.org/xmlns/0.1',
                 'catalog': 'http://anss.org/xmlns/catalog/0.1'}
        cat[0].extra = my_extra.copy()
        # insert a pick with an extra field
        p = Pick()
        p.extra = {'weight': {'value': 2,
                              'namespace': 'http://test.org/xmlns/0.1'}}
        cat[0].picks.append(p)

        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write file
            cat.write(tmpfile, format='QUAKEML', nsmap=nsmap)
            # check contents
            with open(tmpfile, 'rb') as fh:
                # enforce reproducible attribute orders through write_c14n
                obj = etree.fromstring(fh.read()).getroottree()
                buf = io.BytesIO()
                obj.write_c14n(buf)
                buf.seek(0, 0)
                content = buf.read()
            # check namespace definitions in root element
            expected = [b'<q:quakeml',
                        b'xmlns:catalog="http://anss.org/xmlns/catalog/0.1"',
                        b'xmlns:ns0="http://test.org/xmlns/0.1"',
                        b'xmlns:ns1="http://some-page.de/xmlns/1.0"',
                        b'xmlns:q="http://quakeml.org/xmlns/quakeml/1.2"',
                        b'xmlns="http://quakeml.org/xmlns/bed/1.2"']
            for line in expected:
                assert line in content
            # check additional tags
            expected = [
                b'<ns0:custom>True</ns0:custom>',
                b'<ns0:new_tag>1234</ns0:new_tag>',
                b'<ns0:tX>2013-01-02T13:12:14.600000Z</ns0:tX>',
                b'<ns1:public '
                b'another_attrib="another_value" '
                b'some_attrib="some_value">false</ns1:public>'
            ]
            for line in expected:
                assert line in content
            # now, read again to test if it's parsed correctly..
            cat = _read_quakeml(tmpfile)
        # when reading..
        #  - namespace abbreviations should be disregarded
        #  - we always end up with a namespace definition, even if it was
        #    omitted when originally setting the custom tag
        #  - custom namespace abbreviations should attached to Catalog
        assert hasattr(cat[0], 'extra')

        def _tostr(x):
            if isinstance(x, bool):
                if x:
                    return str('true')
                else:
                    return str('false')
            elif isinstance(x, AttribDict):
                for key, value in x.items():
                    x[key].value = _tostr(value['value'])
                return x
            else:
                return str(x)

        for key, value in my_extra.items():
            my_extra[key]['value'] = _tostr(value['value'])
        assert cat[0].extra == my_extra
        assert hasattr(cat[0].picks[0], 'extra')
        assert cat[0].picks[0].extra == \
            {'weight': {'value': '2',
                        'namespace': 'http://test.org/xmlns/0.1'}}
        assert hasattr(cat, 'nsmap')
        assert getattr(cat, 'nsmap')['ns0'] == nsmap['ns0']

    def test_read_same_file_twice_to_same_variable(self):
        """
        Reading the same file twice to the same variable should not raise a
        warning.
        """
        sio = io.BytesIO(b"""<?xml version='1.0' encoding='utf-8'?>
        <q:quakeml xmlns:q="http://quakeml.org/xmlns/quakeml/1.2"
                   xmlns="http://quakeml.org/xmlns/bed/1.2">
          <eventParameters publicID="smi:local/catalog">
            <event publicID="smi:local/event">
              <origin publicID="smi:local/origin">
                <time>
                  <value>1970-01-01T00:00:00.000000Z</value>
                </time>
                <latitude>
                  <value>0.0</value>
                </latitude>
                <longitude>
                  <value>0.0</value>
                </longitude>
                <depth>
                  <value>0.0</value>
                </depth>
                <arrival publicID="smi:local/arrival">
                  <pickID>smi:local/pick</pickID>
                  <phase>P</phase>
                </arrival>
              </origin>
            </event>
          </eventParameters>
        </q:quakeml>
        """)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat1 = read_events(sio)  # NOQA
            cat2 = read_events(sio)  # NOQA

        assert_no_extras(cat1)
        # No warning should have been raised.
        assert len(w) == 0

    def test_focal_mechanism_write_read(self):
        """
        Test for a bug in reading a FocalMechanism without MomentTensor from
        QuakeML file. Makes sure that FocalMechanism.moment_tensor stays None
        if no MomentTensor is in the file.
        """
        memfile = io.BytesIO()
        # create virtually empty FocalMechanism
        fm = FocalMechanism()
        event = Event(focal_mechanisms=[fm])
        cat = Catalog(events=[event])
        cat.write(memfile, format="QUAKEML", validate=True)
        # now read again, and make sure there's no stub MomentTensor, but
        # rather `None`
        memfile.seek(0)
        cat = read_events(memfile, format="QUAKEML")
        assert_no_extras(cat)
        assert cat[0].focal_mechanisms[0].moment_tensor is None

    def test_avoid_empty_stub_elements(self):
        """
        Test for a bug in reading QuakeML. Makes sure that some subelements do
        not get assigned stub elements, but rather stay None.
        """
        # Test 1: Test subelements of moment_tensor
        memfile = io.BytesIO()
        # create virtually empty FocalMechanism
        mt = MomentTensor(derived_origin_id='smi:local/abc')
        fm = FocalMechanism(moment_tensor=mt)
        event = Event(focal_mechanisms=[fm])
        cat = Catalog(events=[event])
        cat.write(memfile, format="QUAKEML", validate=True)
        # now read again, and make sure there's no stub subelements on
        # MomentTensor, but rather `None`
        memfile.seek(0)
        cat = read_events(memfile, format="QUAKEML")
        assert_no_extras(cat)
        assert cat[0].focal_mechanisms[0].moment_tensor.tensor is None
        assert \
            cat[0].focal_mechanisms[0].moment_tensor.source_time_function \
            is None
        # Test 2: Test subelements of focal_mechanism
        memfile = io.BytesIO()
        # create virtually empty FocalMechanism
        fm = FocalMechanism()
        event = Event(focal_mechanisms=[fm])
        cat = Catalog(events=[event])
        cat.write(memfile, format="QUAKEML", validate=True)
        # now read again, and make sure there's no stub MomentTensor, but
        # rather `None`
        memfile.seek(0)
        cat = read_events(memfile, format="QUAKEML")
        assert_no_extras(cat)
        assert cat[0].focal_mechanisms[0].nodal_planes is None
        assert cat[0].focal_mechanisms[0].principal_axes is None

    def test_writing_invalid_quakeml_id(self, testdata):
        """
        Some ids might be invalid. We still want to write them to not mess
        with any external tools relying on the ids. But we also raise a
        warning of course.
        """
        filename = testdata['invalid_id.xml']
        cat = read_events(filename)
        assert cat[0].resource_id.id == \
            "smi:org.gfz-potsdam.de/geofon/RMHP(60)>>ITAPER(3)>>BW(4,5,15)"
        with NamedTemporaryFile() as tf:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                cat.write(tf.name, format="quakeml")
                cat2 = read_events(tf.name)
        assert len(w) == 19
        assert w[0].message.args[0] == \
            ("'smi:org.gfz-potsdam.de/geofon/RMHP(60)>>ITAPER(3)>>BW(4,5,15)' "
             "is not a valid QuakeML URI. It will be in the final file but "
             "note that the file will not be a valid QuakeML file.")
        assert cat2[0].resource_id.id == \
            "smi:org.gfz-potsdam.de/geofon/RMHP(60)>>ITAPER(3)>>BW(4,5,15)"

    def test_reading_invalid_enums(self, testdata):
        """
        Raise a warning when an invalid enum value is attempted to be read.
        """
        filename = testdata['invalid_enum.xml']
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cat = read_events(filename)
        assert len(w) == 1
        assert w[0].message.args[0] == \
            ('Setting attribute "depth_type" failed. Value "randomized" could '
             'not be converted to type "Enum(["from location", "from moment '
             'tensor inversion", ..., "operator assigned", "other"])". The '
             'attribute "depth_type" will not be set and will be missing in '
             'the resulting object.')
        # It should of course not be set.
        assert cat[0].origins[0].depth_type is None

    def test_issue_2339(self):
        """
        Make sure an empty EventDescription object does not prevent a catalog
        from being saved to disk and re-read, while still being equal.
        """
        # create a catalog  with an empty event description
        empty_description = EventDescription()
        cat1 = Catalog(events=[read_events()[0]])
        cat1[0].event_descriptions.append(empty_description)
        # serialize the catalog using quakeml and re-read
        bio = io.BytesIO()
        cat1.write(bio, 'quakeml')
        bio.seek(0)
        cat2 = read_events(bio)
        assert_no_extras(cat2)
        # the text of the empty EventDescription instances should be equal
        text1 = cat1[0].event_descriptions[-1].text
        text2 = cat2[0].event_descriptions[-1].text
        assert text1 == text2
        # the two catalogs should be equal
        assert cat1 == cat2

    def test_native_namespace_in_extra(self):
        """
        Make sure that QuakeML tags that are not the same as the document
        root's namespaces still are handled as custom tags (coming
        after any expected/mandatory tags) and get parsed into extras section
        properly.
        """
        custom1 = {
            'value': u'11111',
            'namespace': 'http://quakeml.org/xmlns/bed/9.99'}
        custom2 = {
            'value': u'22222',
            'namespace': 'http://quakeml.org/xmlns/quakeml/8.87'}
        extra = {'custom1': custom1, 'custom2': custom2}

        cat = Catalog()
        cat.extra = extra

        with io.BytesIO() as buf:
            cat.write(buf, format='QUAKEML')
            buf.seek(0)
            cat2 = read_events(buf, format='QUAKEML')

        assert extra == cat2.extra
        assert ('custom1', custom1) in cat2.extra.items()
        assert ('custom2', custom2) in cat2.extra.items()



# -*- coding: utf-8 -*-
import warnings

from obspy import read_events, read_inventory, UTCDateTime
from obspy.core.event import Catalog, Event, Origin, Pick, WaveformStreamID
from obspy.core.util import NamedTemporaryFile
from obspy.io.hypodd import pha
import pytest


class TestPHA():
    """
    Test suite for obspy.io.hypodd.pha
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        self.fname = testdata['example.pha']
        self.fname2 = testdata['60s_nan.pha']

    def test_is_pha(self):
        assert pha._is_pha(self.fname)

    def test_is_not_pha(self, datapath):
        fname = datapath.parent / 'test_pha.py'
        assert not pha._is_pha(fname)

    def test_read_pha(self):
        cat = read_events(self.fname)
        assert len(cat) == 2
        event = cat[0]
        assert len(event.origins) == 1
        assert len(event.magnitudes) == 1
        ori = event.preferred_origin()
        mag = event.preferred_magnitude()
        assert str(ori.time) == '2025-05-14T14:35:35.510000Z'
        assert ori.latitude == 40.2254
        assert ori.longitude == 10.4496
        assert ori.depth == 9408.0
        assert mag.mag == 3.50
        assert len(event.picks) == 2
        assert len(ori.arrivals) == 2
        target = {'FUR': (3.52199909, 1.0000, 'P'),
                  'WET': (5.86199909, 1.0000, 'S')}
        for arr in ori.arrivals:
            pick = arr.pick_id.get_referred_object()
            sta = pick.waveform_id.station_code
            reltime = pick.time - ori.time
            assert round(abs(reltime-target[sta][0]), 4) == 0
            assert arr.time_weight == target[sta][1]
            assert arr.phase == target[sta][2]
            assert arr.phase == pick.phase_hint
        event = cat[1]
        ori = event.preferred_origin()
        assert round(abs(ori.latitude_errors.uncertainty-0.0045), 5) == 0
        assert ori.latitude_errors.uncertainty < \
            ori.longitude_errors.uncertainty
        assert ori.depth_errors.uncertainty == 400
        assert ori.quality.standard_error == 0.3
        assert ori.quality.associated_phase_count == 2

    def test_60s_nan(self):
        """
        issue 2627
        """
        cat = read_events(self.fname2)
        event = cat[0]
        assert len(event.origins) == 1
        assert len(event.magnitudes) == 1
        ori = event.preferred_origin()
        assert str(ori.time) == '2025-05-14T14:36:00.000000Z'
        event = cat[1]
        assert len(event.origins) == 1
        assert len(event.magnitudes) == 0

    def test_populate_waveform_id(self):
        inv = read_inventory()
        with warnings.catch_warnings(record=True) as ws:
            warnings.resetwarnings()
            cat = read_events(self.fname, inventory=inv)
            assert len(ws) == 4
        # FUR with channels, WET without channels
        inv = read_inventory().select(channel='HH?')
        inv[0][1].channels = []
        cat = read_events(self.fname, inventory=inv,
                          default_seedid='BLA.{}.11.DH{}',
                          seedid_map={'UBR': 'BLB.{}.00.BH{}'}, warn=False)
        assert len(cat) == 2
        picks = cat[0].picks + cat[1].picks
        assert len(picks) == 4
        waveform_ids = [p.waveform_id.get_seed_string() for p in picks]
        assert 'GR.FUR..HHZ' in waveform_ids
        assert 'BLA.WET.11.DHN' in waveform_ids
        assert 'BLB.UBR.00.BHZ' in waveform_ids
        assert 'BLA.WERD.11.DH' in waveform_ids

    def test_eventid_map(self):
        cat = read_events(self.fname) + read_events(self.fname2)
        cat[0].resource_id = 'X'
        cat[1].resource_id = '2f'
        cat[2].resource_id = 'Y'
        cat[3].resource_id = '1234567890Z'
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            with pytest.warns(UserWarning, match='Missing mag'):
                eventid_map = cat.write(tempfile, 'HYPODDPHA')
            cat2 = read_events(tempfile, eventid_map=eventid_map)
            cat3 = read_events(tempfile)
            with pytest.warns(UserWarning, match='Missing mag'):
                eventid_map2 = cat.write(tempfile, 'HYPODDPHA',
                                         eventid_map=eventid_map)
        assert cat2[0].resource_id.id.split('/')[-1] == 'X'
        assert cat2[1].resource_id.id.split('/')[-1] == '2f'
        assert cat2[2].resource_id.id.split('/')[-1] == 'Y'
        assert cat2[3].resource_id.id.split('/')[-1] == '1234567890Z'
        assert cat3[0].resource_id.id.split('/')[-1] == '1'
        assert cat3[1].resource_id.id.split('/')[-1] == '2'
        assert cat3[2].resource_id.id.split('/')[-1] == '3'
        assert cat3[3].resource_id.id.split('/')[-1] == '123456789'
        assert eventid_map2 == eventid_map

    def test_write_pha(self):
        with open(self.fname) as f:
            filedata = f.read()
        cat = read_events(self.fname)
        cat[0].origins[0].arrivals[0].time_weight = None
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            cat.write(tempfile, 'HYPODDPHA')
            with open(tempfile) as f:
                filedata2 = f.read()
        assert filedata2.replace(' ', '') == filedata.replace(' ', '')

    def test_write_pha_minimal(self):
        ori = Origin(time=UTCDateTime(0), latitude=42, longitude=43,
                     depth=10000)
        pick = Pick(time=UTCDateTime(10), phase_hint='S',
                    waveform_id=WaveformStreamID(station_code='STA'))
        del ori.latitude_errors
        del ori.longitude_errors
        del ori.depth_errors
        cat = Catalog([Event(origins=[ori], picks=[pick])])
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            with pytest.warns(UserWarning, match='Missing mag'):
                cat.write(tempfile, 'HYPODDPHA')
            cat2 = read_events(tempfile)
        assert len(cat2) == 1
        assert len(cat2[0].picks) == 1



# -*- coding: utf-8 -*-
"""
obspy.io.nordic - Nordic file format support for ObsPy
======================================================

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
class NordicParsingError(Exception):
    """
    Internal general error for IO operations in obspy.core.io.nordic.
    """
    def __init__(self, value):
        self.value = value



from obspy.taup import TauPyModel

model = TauPyModel(model='iasp91')
arrivals = model.get_ray_paths(500, 140, phase_list=['PP', 'SSS'])
arrivals.plot_rays(plot_type='cartesian', phase_list=['PP', 'SSS'],
                   plot_all=False, legend=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
obspy.io.stationtxt - FDSNWS station text file read and write support for ObsPy
===============================================================================

This module provides read support for the station text files served by the FDSN
web services.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


Example
-------

Don't use this module directly but utilize it through the
:func:`~obspy.core.inventory.inventory.read_inventory` function.

>>> import obspy
>>> inv = obspy.read_inventory("/path/to/channel_level_fdsn.txt")
>>> print(inv)  # # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
Inventory created at ...
    Created by: ObsPy ...
            https://www.obspy.org
    Sending institution: None
    Contains:
        Networks (2):
            AK, AZ
        Stations (3):
            AK.BAGL ()
            AK.BWN ()
            AZ.BZN ()
        Channels (6):
            AK.BAGL..LHZ, AK.BWN..LHZ (2x), AZ.BZN..LHZ (3x)
"""
if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
import io

from obspy.core.event import read_events
from obspy.io.nied.fnetmt import _is_fnetmt_catalog


class TestFNETMTCatalogReading():
    """
    Test everything related to reading an F-net moment tensor catalog.
    """
    def test_read_fnetmt_catalog(self, testdata):
        testfile = testdata['FNETMTCATALOG']
        cat = read_events(testfile, 'FNETMT')
        assert len(cat) == 1
        ev = cat[0]
        assert len(ev.origins) == 2
        assert len(ev.magnitudes) == 2

    def test_read_fnetmt_catalog_from_open_files(self, testdata):
        """
        Tests that reading an F-net moment tensor file from an open file works.
        """
        testfile = testdata['FNETMTCATALOG']
        with open(testfile, "rb") as fh:
            read_events(fh)

    def test_read_fnetmt_catalog_from_bytes_io(self, testdata):
        """
        Tests that reading an F-net moment tensor file from a BytesIO objects
        works.
        """
        testfile = testdata['FNETMTCATALOG']
        with open(testfile, "rb") as fh:
            buf = io.BytesIO(fh.read())

        with buf:
            read_events(buf)

    def test_is_fnetmt_catalog(self, testdata, datapath):
        """
        This tests the _is_fnetmt_catalog method by validating that each file
        in the data directory is an F-net catalog file and each file in the
        working directory is not.

        The filenames are hard coded so the test will not fail with future
        changes in the structure of the package.
        """
        # F-net catalog file names.
        fnetmt_filenames = ['FNETMTCATALOG']

        # Non F-net file names.
        non_fnetmt_filenames = ['test_fnetmt_reading.py',
                                '__init__.py']
        # Loop over F-net files
        for _i in fnetmt_filenames:
            filename = testdata[_i]
            is_fnetmt = _is_fnetmt_catalog(filename)
            assert is_fnetmt
        # Loop over non F-net files
        for _i in non_fnetmt_filenames:
            filename = datapath.parent / _i
            is_fnetmt = _is_fnetmt_catalog(filename)
            assert not is_fnetmt



# -*- coding: utf-8 -*-


def parse_wtn_frame_to_alsep_words(frame):
    alsep_words = [None] * 65
    for i in range(21):
        j = i * 4 + 8
        alsep_words[i * 3 + 1] = (frame[j] << 2) | (frame[j + 1] >> 6)
        alsep_words[i * 3 + 2] = \
            ((frame[j + 1] & 0x1f) << 5) | ((frame[j + 2] >> 3) & 0x1f)
        alsep_words[i * 3 + 3] = ((frame[j + 2] & 0x03) << 8) | (frame[j + 3])
    alsep_words[64] = (frame[92] << 2) | (frame[93] >> 6)
    return alsep_words


def parse_wth_frame_to_geophone_data(frame):
    geophone = {1: [0] * 20, 2: [0] * 20, 3: [0] * 20, 4: [0] * 20,
                'status': [0] * 20}

    part1 = (frame[9] >> 1) & 0x1f
    part1 <<= 3
    geophone[1][0] = part1

    part2 = frame[9] & 0x01
    part2 = (part2 << 4) | ((frame[10] >> 4) & 0x0f)
    part2 <<= 3
    geophone[2][0] = part2

    part3 = frame[10] & 0x0f
    part3 = (part3 << 1) | (frame[11] >> 7)
    part3 <<= 3
    geophone[3][0] = part3

    part4 = (frame[11] >> 2) & 0x1f
    part4 <<= 3
    geophone[4][0] = part4

    geophone['status'][0] = -1

    for i in range(1, 20):
        j = (i + 2) * 4

        part1 = (frame[j] >> 1) << 1
        geophone[1][i] = part1

        part2 = frame[j] & 0x01
        part2 = (part2 << 6) | (frame[j + 1] >> 2)
        part2 <<= 1
        geophone[2][i] = part2

        part3 = frame[j + 1] & 0x03
        part3 = (part3 << 5) | (frame[j + 2] >> 3)
        part3 <<= 1
        geophone[3][i] = part3

        part4 = frame[j + 2] & 0x07
        part4 = (part4 << 4) | (frame[j + 3] >> 4)
        part4 <<= 1
        geophone[4][i] = part4

        geophone['status'][i] = (frame[j + 3] >> 2) & 0x03

    sub_frame_array = [-1, 2, 3, 1]
    geophone['sub_frame'] = sub_frame_array[geophone['status'][19]]

    return geophone



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import collections
import warnings
from unittest import mock

import pytest

import obspy
from obspy.clients.fdsn.header import FDSNNoDataException
from obspy.clients.fdsn.routing.routing_client import (
    BaseRoutingClient, RoutingClient)
from obspy.clients.fdsn.routing.eidaws_routing_client import (
    EIDAWSRoutingClient)
from obspy.clients.fdsn.routing.federator_routing_client import (
    FederatorRoutingClient)


_DummyResponse = collections.namedtuple("_DummyResponse", ["content"])


class TestBaseRoutingClient():
    @classmethod
    def setup_class(cls):
        # Need to inherit to add two required method by the abstract base
        # class.
        class _DummyBaseRoutingClient(BaseRoutingClient):  # pragma: no cover
            def get_service_version(self):
                """
                Return a semantic version number of the remote service as a
                string.
                """
                return "0.0.0"

            def _handle_requests_http_error(self, r):
                raise NotImplementedError

        cls._cls_object = _DummyBaseRoutingClient
        cls._cls = ("obspy.clients.fdsn.routing.routing_client."
                    "BaseRoutingClient")

    def test_router_intialization_helper_function(self):
        c = RoutingClient("eida-routing")
        assert isinstance(c, EIDAWSRoutingClient)

        c = RoutingClient("earthscope-federator")
        assert isinstance(c, FederatorRoutingClient)

        msg = "Routing type 'random' is not implemented. Available types: " \
              "`earthscope-federator`, `eida-routing`"
        with pytest.raises(NotImplementedError, match=msg):
            RoutingClient("random")

    def test_expansion_of_include_and_exclude_providers(self):
        c = self._cls_object(
            include_providers=["EARTHSCOPE", "http://example.com"],
            exclude_providers=["BGR", "http://example2.com"])
        assert c.include_providers == ["service.earthscope.org", "example.com"]
        assert c.exclude_providers == ["eida.bgr.de", "example2.com"]

        # None are set.
        c = self._cls_object()
        assert c.include_providers == []
        assert c.exclude_providers == []

        # Single strings.
        c = self._cls_object(include_providers="EARTHSCOPE",
                             exclude_providers="BGR")
        assert c.include_providers == ["service.earthscope.org"]
        assert c.exclude_providers == ["eida.bgr.de"]

        c = self._cls_object(include_providers="http://example.com/path",
                             exclude_providers="http://example2.com")
        assert c.include_providers == ["example.com/path"]
        assert c.exclude_providers == ["example2.com"]

    def test_request_filtering(self):
        split = {
            # Note that this is HTTPS.
            "https://example.com": "1234",
            "http://example2.com": "1234",
            "http://example3.com": "1234",
            "https://service.earthscope.org": "1234"
        }

        c = self._cls_object(include_providers=["EARTHSCOPE",
                                                "http://example.com"])
        assert c._filter_requests(split) == {
            "https://example.com": "1234",
            "https://service.earthscope.org": "1234"
        }

        c = self._cls_object(exclude_providers=["EARTHSCOPE",
                                                "http://example.com"])
        assert c._filter_requests(split) == {
            "http://example2.com": "1234",
            "http://example3.com": "1234"
        }

        # Both filters are always applied - it might result in zero
        # remaining providers.
        c = self._cls_object(
            include_providers=["EARTHSCOPE", "http://example.com"],
            exclude_providers=["EARTHSCOPE", "http://example.com"])
        assert c._filter_requests(split) == {}

    def test_downloading_waveforms(self):
        split = {
            "https://example.com": "1234",
            "http://example2.com": "1234",
            "http://example3.com": "1234",
            "https://service.earthscope.org": "1234"
        }
        with mock.patch("obspy.clients.fdsn.client.Client") as p:
            mock_instance = p.return_value
            mock_instance.get_waveforms_bulk.return_value = obspy.read()
            # Only accept test1 as a an argument.
            mock_instance.services = {"dataselect": {"test1": True}}
            c = self._cls_object(debug=False, timeout=240)
            # test2 should not be passed on.
            st = c._download_waveforms(split=split, test1="a", test2="b")

        assert len(st) == 12
        # Test initialization.
        assert p.call_count == 4
        assert {_i[0][0] for _i in p.call_args_list} == {*split.keys()}
        assert {_i[1]["debug"] for _i in p.call_args_list} == {False}
        assert {_i[1]["timeout"] for _i in p.call_args_list} == {240}

        # Waveform download.
        wf_bulk = mock_instance.get_waveforms_bulk
        assert wf_bulk.call_count == 4
        assert {_i[0][0] for _i in wf_bulk.call_args_list} == {"test1=a\n1234"}
        for _i in wf_bulk.call_args_list:
            assert _i[1] == {}

        # Once again, but raising exceptions this time.
        with mock.patch("obspy.clients.fdsn.client.Client") as p:
            mock_instance = p.return_value
            mock_instance.get_waveforms_bulk.side_effect = \
                FDSNNoDataException("No data")
            # Only accept test1 as a an argument.
            mock_instance.services = {"dataselect": {"test1": True}}
            c = self._cls_object(debug=False, timeout=240)
            # test2 should not be passed on.
            st = c._download_waveforms(split=split, test1="a", test2="b")

        assert len(st) == 0

        # Provider filtering might result in no data left.
        c.include_providers = "http://random.com"
        msg = "Nothing remains to download after the provider " \
              "inclusion/exclusion filters have been applied."
        with pytest.raises(FDSNNoDataException, match=msg):
            c._download_waveforms(split=split, test1="a", test2="b")

    def test_downloading_stations(self):
        split = {
            "https://example.com": "1234",
            "http://example2.com": "1234",
            "http://example3.com": "1234",
            "https://service.earthscope.org": "1234"
        }
        with mock.patch("obspy.clients.fdsn.client.Client") as p:
            mock_instance = p.return_value
            mock_instance.get_stations_bulk.return_value = \
                obspy.Inventory([], "")
            # Only accept test1 as a an argument.
            mock_instance.services = {"station": {"test1": True}}
            c = self._cls_object(debug=False, timeout=240)
            # test2 should not be passed on.
            c._download_stations(split=split, test1="a", test2="b")

        # Test initialization.
        assert p.call_count == 4
        assert {_i[0][0] for _i in p.call_args_list} == {*split.keys()}
        assert {_i[1]["debug"] for _i in p.call_args_list} == {False}
        assert {_i[1]["timeout"] for _i in p.call_args_list} == {240}

        # Station download.
        wf_bulk = mock_instance.get_stations_bulk
        assert wf_bulk.call_count == 4
        assert {_i[0][0] for _i in wf_bulk.call_args_list} == {"test1=a\n1234"}
        for _i in wf_bulk.call_args_list:
            assert _i[1] == {}

    def test_unexpected_exception_handling(self):
        split = {"https://example.com": "1234"}

        with mock.patch("obspy.clients.fdsn.client.Client") as p:
            mock_instance = p.return_value
            mock_instance.get_stations_bulk.side_effect = ValueError("random")
            c = self._cls_object(debug=False, timeout=240)
            # Should not fail, but a warning should be raised.
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                inv = c._download_stations(split=split)

        # Returns an empty inventory.
        assert isinstance(inv, obspy.core.inventory.Inventory)
        assert len(inv) == 0

        # Raises a nice warning.
        assert len(w) == 1
        msg = w[0].message.args[0]
        assert msg.startswith(
            "Failed to download data of type 'station' from "
            "'https://example.com' due to:")
        assert "ValueError: random" in msg



# -*- coding: utf-8 -*-
"""
Helpers to download potentially big data sets across a number of FDSN web
services in an automated fashion.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2014-2015
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import collections
import logging
from multiprocessing.pool import ThreadPool
import os
import warnings

from obspy.clients.fdsn.header import URL_MAPPINGS, FDSNException
from obspy.clients.fdsn.client import Client

from . import utils
from .download_helpers import ClientDownloadHelper, STATUS


logger = logging.getLogger("obspy.clients.fdsn.mass_downloader")


class FDSNMassDownloaderException(FDSNException):
    """
    Base exception raised by the download helpers.
    """
    pass


class MassDownloader(object):
    """
    Class facilitating data acquisition across all FDSN web service
    implementations.

    :param providers: List of FDSN client names or service URLS. Will use
        all FDSN implementations known to ObsPy except RASPISHAKE (generally
        worse quality data) and IRISPH5 (active source / nodal experiments that
        might match a very large amount of data occasionally) if set to None.
        The order in the list also determines their priority, if data is
        available at more then one provider it will always be downloaded from
        the provider that comes first in the list. To include RASPISHAKE and
        IRISPH5, you must set this parameter to
        `obspy.clients.fdsn.header.URL_MAPPINGS` explicitly.
    :param debug: Debug flag passed to the underlying FDSN web service clients.
    :type providers: list[str] or :class:`~obspy.clients.fdsn.client.Client`
        instances
    """
    def __init__(self, providers=None, debug=False, configure_logging=True):
        if configure_logging:
            logger.setLevel(logging.DEBUG)
            # Prevent propagating to higher loggers.
            logger.propagate = 0
            # Console log handler.
            ch = logging.StreamHandler()
            ch.setLevel(logging.INFO)
            # Add formatter
            formatter = logging.Formatter(
                    "[%(asctime)s] - %(name)s - %(levelname)s: %(message)s")
            ch.setFormatter(formatter)
            logger.addHandler(ch)
        self.debug = debug
        # If not given, use all providers ObsPy knows. They will be sorted
        # alphabetically except that ORFEUS is second to last and IRIS last.
        # The reason for this order is that smaller data centers can be
        # expected to have more up-to-date meta information about their own
        # data and thus should be prioritized when downloading data.
        if providers is None:
            providers = dict(URL_MAPPINGS.items())
            _p = []

            if "RASPISHAKE" in providers:
                # exclude RASPISHAKE by default
                del providers["RASPISHAKE"]

            if "IRIS" in providers or "EARTHSCOPE" in providers:
                has_earthscope = True
                if "IRIS" in providers:
                    del providers["IRIS"]
                if "EARTHSCOPE" in providers:
                    del providers["EARTHSCOPE"]
            else:
                has_earthscope = False

            if "ODC" in providers:
                providers["ORFEUS"] = providers["ODC"]
                del providers["ODC"]

            if "ORFEUS" in providers:
                has_orfeus = True
                del providers["ORFEUS"]
            else:
                has_orfeus = False

            # leave out IRISPH5 which is for nodal experiments and might match
            # insanely large datasets, depending on restrictions
            if "IRISPH5" in providers:
                del providers["IRISPH5"]

            _p = sorted(providers)
            if has_orfeus:
                _p.append("ORFEUS")
            if has_earthscope:
                _p.append("EARTHSCOPE")

            providers = _p

        self.providers = tuple(providers)

        # Initialize all clients.
        self._initialized_clients = collections.OrderedDict()
        self._initialize_clients()

    def download(self, domain, restrictions, mseed_storage,
                 stationxml_storage, download_chunk_size_in_mb=20,
                 threads_per_client=3, print_report=True):
        """
        Launch the actual data download.

        :param domain: The download domain.
        :type domain: :class:`~.domain.Domain`
        :param restrictions: Non-spatial downloading restrictions.
        :type restrictions: :class:`~.restrictions.Restrictions`
        :param mseed_storage: Where to store the waveform files. See
            the :mod:`~obspy.clients.fdsn.mass_downloader` documentation for
            more details.
        :type mseed_storage: str or callable
        :param stationxml_storage: Where to store the StationXML files. See
            the :mod:`~obspy.clients.fdsn.mass_downloader` documentation for
            more details.
        :type stationxml_storage: str or callable
        :param download_chunk_size_in_mb: MiniSEED data will be downloaded
            in bulk chunks. This settings limits the chunk size. A higher
            number means that less total download requests will be sent,
            but each individual download request will be larger.
        :type download_chunk_size_in_mb: float
        :param threads_per_client: The number of download threads launched
            per client.
        :type threads_per_client: int
        """
        # The downloads from each client will be handled separately.
        # Nonetheless collect all in this dictionary.
        client_download_helpers = {}

        # Do it sequentially for each client. Doing it in parallel is not
        # really feasible as long as the availability queries are not
        # reliable for all endpoints.
        for client_name, client in self._initialized_clients.items():
            # Log some information about preexisting data.
            station_count = 0
            for _c in client_download_helpers.values():
                station_count += len([
                    _i for _i in _c.stations.values() if
                    (_i.stationxml_status == STATUS.EXISTS) or
                    (_i.has_existing_or_downloaded_time_intervals)])
            logger.info("Total acquired or preexisting stations: %i" %
                        station_count)

            # The client download helper object is responsible for the
            # downloads of a single FDSN endpoint.
            helper = ClientDownloadHelper(
                client=client, client_name=client_name,
                restrictions=restrictions, domain=domain,
                mseed_storage=mseed_storage,
                stationxml_storage=stationxml_storage, logger=logger)
            existing_client_dl_helpers = list(
                client_download_helpers.values())
            client_download_helpers[client_name] = helper

            # Request the availability.
            helper.get_availability()

            # Continue if there is no data.
            if not helper:
                logger.info("Client '%s' - No data available." % client_name)
                continue

            # First filter stage. Remove stations based on the station id,
            # e.g. NETWORK.STATION. Remove all that already exist.
            helper.discard_stations(
                existing_client_dl_helpers=existing_client_dl_helpers)

            # Continue if there is no data.
            if not helper:
                logger.info("Client '%s' - No new data available after "
                            "discarding already downloaded data." %
                            client_name)
                continue

            # If the availability information is reliable, the filtering
            # will happen before the downloading.
            if helper.is_availability_reliable:
                helper.filter_stations_based_on_minimum_distance(
                    existing_client_dl_helpers=existing_client_dl_helpers)
                # Continue if there is no data left after the filtering.
                if not helper:
                    logger.info("Client '%s' - No new data available after "
                                "discarding based on the minimal "
                                "inter-station distance." % client_name)
                    continue

            logger.info("Client '%s' - Will attempt to download data from %i "
                        "stations." % (client_name, len(helper)))

            # Download MiniSEED data.
            helper.prepare_mseed_download()
            helper.download_mseed(chunk_size_in_mb=download_chunk_size_in_mb,
                                  threads_per_client=threads_per_client)

            # Download StationXML data.
            helper.prepare_stationxml_download()
            helper.download_stationxml(threads=threads_per_client)

            # Sanitize the downloaded things if desired. Assures that all
            # waveform data also has the corresponding station information.
            if restrictions.sanitize:
                helper.sanitize_downloads()

            if not helper:
                logger.info("Client '%s' - No data could be downloaded." %
                            client_name)
                continue

            # Filter afterwards if availability information is not reliable.
            # This unfortunately results in already downloaded data being
            # discarded but it is the only currently feasible way.
            if not helper.is_availability_reliable:
                helper.filter_stations_based_on_minimum_distance(
                    existing_client_dl_helpers=existing_client_dl_helpers)

        if print_report:
            # Collect already existing things.
            existing_miniseed_files = []
            existing_stationxml_files = []
            new_miniseed_files = collections.defaultdict(list)
            new_stationxml_files = collections.defaultdict(list)

            for cdh in client_download_helpers.values():
                for station in cdh.stations.values():
                    if station.stationxml_status == STATUS.EXISTS:
                        existing_stationxml_files.append(
                            station.stationxml_filename)
                    elif station.stationxml_status == STATUS.DOWNLOADED:
                        new_stationxml_files[cdh.client_name].append(
                            station.stationxml_filename)
                    for channel in station.channels:
                        for ti in channel.intervals:
                            if ti.status == STATUS.EXISTS:
                                existing_miniseed_files.append(ti.filename)
                            elif ti.status == STATUS.DOWNLOADED:
                                new_miniseed_files[cdh.client_name].append(
                                    ti.filename)

            def count_filesize(list_of_files):
                return sum([os.path.getsize(_i) for _i in list_of_files if
                            os.path.exists(_i)])

            logger.info(30 * "=" + " Final report")
            logger.info("%i MiniSEED files [%.1f MB] already existed." % (
                len(existing_miniseed_files),
                count_filesize(existing_miniseed_files) / 1024.0 ** 2))
            logger.info("%i StationXML files [%.1f MB] already existed." % (
                len(existing_stationxml_files),
                count_filesize(existing_stationxml_files) / 1024.0 ** 2))

            total_downloaded_filesize = 0
            for cdh in client_download_helpers.values():
                mseed_files = new_miniseed_files[cdh.client_name]
                stationxml_files = new_stationxml_files[cdh.client_name]
                mseed_filesize = count_filesize(mseed_files)
                stationxml_filesize = count_filesize(stationxml_files)
                total_downloaded_filesize += mseed_filesize
                total_downloaded_filesize += stationxml_filesize
                logger.info("Client '%s' - Acquired %i MiniSEED files "
                            "[%.1f MB]." % (cdh.client_name, len(mseed_files),
                                            mseed_filesize / 1024.0 ** 2))
                logger.info("Client '%s' - Acquired %i StationXML files "
                            "[%.1f MB]." % (
                                cdh.client_name, len(stationxml_files),
                                stationxml_filesize / 1024.0 ** 2))
            logger.info("Downloaded %.1f MB in total." % (
                total_downloaded_filesize / 1024.0 ** 2))

        return client_download_helpers

    def _initialize_clients(self):
        """
        Initialize all clients.
        """
        logger.info("Initializing FDSN client(s) for %s." % ", ".join(
            _i.base_url if hasattr(_i, "base_url") else _i
            for _i in self.providers))

        def _get_client(client_name):
            # It might already be an initialized client - in that case just
            # use it.
            if isinstance(client_name, Client):
                name, client = client_name.base_url, client_name
            else:
                try:
                    this_client = Client(client_name, debug=self.debug)
                    name, client = client_name, this_client
                except utils.ERRORS as e:
                    if "timeout" in str(e).lower():
                        extra = " (timeout)"
                    else:
                        extra = ""
                    logger.warn("Failed to initialize client '%s'.%s"
                                % (client_name, extra))
                    return client_name, None

            services = sorted([_i for _i in client.services.keys()
                               if not _i.startswith("available")])
            if "dataselect" not in services or "station" not in services:
                logger.info("Cannot use client '%s' as it does not have "
                            "'dataselect' and/or 'station' services."
                            % name)
                return name, None
            return name, client

        # Catch warnings in the main thread. The catch_warnings() context
        # manager does not reliably work when used in multiple threads.
        p = ThreadPool(len(self.providers))
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            clients = p.map(_get_client, self.providers)
        p.close()
        for warning in w:
            logger.debug("Warning during initialization of one of the "
                         "clients: " + str(warning.message))

        clients = {key: value for key, value in clients if value is not None}

        # Write to initialized clients dictionary preserving order. Remember
        # that each passed provider might already be an initialized client
        # instance.
        for client in self.providers:
            if client not in clients and client not in clients.values():
                continue
            name = client.base_url if hasattr(client, "base_url") else client
            self._initialized_clients[name] = clients[name]

        logger.info("Successfully initialized %i client(s): %s."
                    % (len(self._initialized_clients),
                       ", ".join(self._initialized_clients.keys())))


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.clients.neic - CWB query module for ObsPy
===============================================
The obspy.clients.neic package contains a client for the NEIC CWB Query server.
A public server is at 137.227.230.97" for cwbpub.usgs.gov on port 2061.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from .client import Client  # NOQA


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The audio wav.core test suite.
"""
import os

import numpy as np
import pytest

from obspy import Stream, Trace, read
from obspy.core.util import NamedTemporaryFile
from obspy.io.wav.core import WIDTH2DTYPE


class TestCore():
    """
    Test cases for audio WAV support
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        self.file = testdata['3cssan.near.8.1.RNON.wav']

    def test_read_via_obspy(self):
        """
        Read files via obspy.core.Trace
        """
        testdata = np.array([64, 78, 99, 119, 123, 107,
                             72, 31, 2, 0, 30, 84, 141])
        tr = read(self.file)[0]
        assert tr.stats.npts == 2599
        assert tr.stats['sampling_rate'] == 7000
        np.testing.assert_array_equal(tr.data[:13], testdata)
        tr2 = read(self.file, format='WAV')[0]
        assert tr2.stats.npts == 2599
        assert tr2.stats['sampling_rate'] == 7000
        np.testing.assert_array_equal(tr.data[:13], testdata)

    def test_read_head_via_obspy(self):
        """
        Read files via obspy.core.Trace
        """
        tr = read(self.file, headonly=True)[0]
        assert tr.stats.npts == 2599
        assert tr.stats['sampling_rate'] == 7000
        assert str(tr.data) == '[]'

    def test_read_and_write_via_obspy(self, testdata):
        """
        Read and Write files via obspy.core.Trace
        """
        data = np.array([111, 111, 111, 111, 111, 109, 106, 103, 103,
                         110, 121, 132, 139])
        with NamedTemporaryFile() as fh:
            testfile = fh.name
            self.file = testdata['3cssan.reg.8.1.RNON.wav']
            tr = read(self.file, format='WAV')[0]
            assert tr.stats.npts == 10599
            assert tr.stats['sampling_rate'] == 7000
            np.testing.assert_array_equal(tr.data[:13], data)
            # write
            st2 = Stream()
            st2.traces.append(Trace())
            st2[0].data = tr.data.copy()  # copy the data
            st2.write(testfile, format='WAV', framerate=7000)
            # read without giving the WAV format option
            tr3 = read(testfile)[0]
            assert tr3.stats == tr.stats
            np.testing.assert_array_equal(tr3.data[:13], data)

    def test_rescale_on_write(self, testdata):
        """
        Read and Write files via obspy.core.Trace
        """
        with NamedTemporaryFile() as fh:
            testfile = fh.name
            self.file = testdata['3cssan.reg.8.1.RNON.wav']
            tr = read(self.file, format='WAV')[0]
            for width in (1, 2, 4, None):
                tr.write(testfile, format='WAV', framerate=7000, width=width,
                         rescale=True)
                if width is not None:
                    tr2 = read(testfile, format='WAV')[0]
                    maxint = 2 ** (8 * width - 1) - 1
                    dtype = WIDTH2DTYPE[width]
                    assert maxint == abs(tr2.data).max()
                    expected = (tr.data / abs(tr.data).max() *
                                maxint).astype(dtype)
                    np.testing.assert_array_almost_equal(tr2.data, expected)

    def test_write_stream_via_obspy(self, testdata):
        """
        Write streams, i.e. multiple files via obspy.core.Trace
        """
        data = np.array([111, 111, 111, 111, 111, 109, 106, 103, 103,
                         110, 121, 132, 139])
        with NamedTemporaryFile() as fh:
            testfile = fh.name
            self.file = testdata['3cssan.reg.8.1.RNON.wav']
            tr = read(self.file, format='WAV')[0]
            np.testing.assert_array_equal(tr.data[:13], data)
            # write
            st2 = Stream([Trace(), Trace()])
            st2[0].data = tr.data.copy()       # copy the data
            st2[1].data = tr.data.copy() // 2  # be sure data are different
            st2.write(testfile, format='WAV', framerate=7000)
            try:
                # read without giving the WAV format option
                base, ext = os.path.splitext(testfile)
                testfile0 = "%s%03d%s" % (base, 0, ext)
                testfile1 = "%s%03d%s" % (base, 1, ext)
                tr30 = read(testfile0)[0]
                tr31 = read(testfile1)[0]
                assert tr30.stats == tr.stats
                assert tr31.stats == tr.stats
                np.testing.assert_array_equal(tr30.data[:13], data)
                np.testing.assert_array_equal(tr31.data[:13], data // 2)
            finally:
                os.remove(testfile0)
                os.remove(testfile1)
